""" Backend access class. """
from functools import partial

from elasticsearch.exceptions import NotFoundError, TransportError
from pymongo import UpdateOne

from biothings import config as btconfig
from biothings.utils.es import ESIndexer


# Generic base backend
class DocBackendBase(object):
    name = "Undefined"

    def prepare(self):
        """if needed, add extra preparation steps here."""

    @property
    def target_name(self):
        raise NotImplementedError()

    @property
    def version(self):
        raise NotImplementedError()

    def insert(self, doc_li):
        raise NotImplementedError()

    def update(self, id, extra_doc):
        """update only, no upsert."""
        raise NotImplementedError()

    def drop(self):
        raise NotImplementedError()

    def get_id_list(self):
        raise NotImplementedError()

    def get_from_id(self, id):
        raise NotImplementedError()

    def finalize(self):
        """if needed, for example for bulk updates, perform flush
        at the end of updating.
        Final optimization or compacting can be done here as well.
        """


class DocMemoryBackend(DocBackendBase):
    name = "memory"

    def __init__(self, target_name=None):
        """target_dict is None or a dict."""
        self.target_dict = {}
        self._target_name = target_name or "unnamed"

    @property
    def target_name(self):
        return self._target_name

    def insert(self, doc_li):
        for doc in doc_li:
            self.target_dict[doc["_id"]] = doc

    def update(self, id, extra_doc):
        current_doc = self.target_dict.get(id, None)
        if current_doc:
            current_doc.update(extra_doc)
            self.target_dict[id] = current_doc

    def drop(self):
        self.target_dict = {}

    def get_id_list(self):
        return self.target_dict.keys()

    def get_from_id(self, id):
        return self.target_dict[id]

    def finalize(self):
        """dump target_dict into a file."""
        from biothings.utils.common import dump

        dump(self.target_dict, self.target_name + ".pyobj")


class DocMongoBackend(DocBackendBase):
    name = "mongo"

    def __init__(self, target_db, target_collection=None):
        """target_collection is a pymongo collection object."""
        if callable(target_db):
            self._target_db_provider = target_db
            self._target_db = None
        else:
            self._target_db = target_db
        if target_collection:
            self.target_collection = target_collection

    def __eq__(self, other):
        if not isinstance(other, DocMongoBackend):
            return False
        return (
            self.target_name == other.target_name
            and self.target_collection.database.name == other.target_collection.database.name
            and self.target_collection.database.client.address == other.target_collection.database.client.address
        )

    @property
    def target_name(self):
        return self.target_collection.name

    @property
    def version(self):
        import biothings.utils.mongo as mongo

        if self.target_collection.database.name == btconfig.DATA_SRC_DATABASE:
            fulln = mongo.get_source_fullname(self.target_collection.name)
            if not fulln:
                return
            mainsrc = fulln.split(".")[0]
            col = mongo.get_src_dump()
            src = col.find_one({"_id": mainsrc})
            return src.get("release")
        elif self.target_collection.database.name == btconfig.DATA_TARGET_DATABASE:
            col = mongo.get_src_build()
            tgt = col.find_one({"_id": self.target_collection.name})
            if not tgt:
                return
            return tgt.get("_meta", {}).get("build_version")
        else:
            return None

    @property
    def target_db(self):
        if self._target_db is None:
            self._target_db = self._target_db_provider()
        return self._target_db

    def count(self):
        return self.target_collection.count()

    def insert(self, docs):
        try:
            res = self.target_collection.insert_many(documents=docs)
            return len(res.inserted_ids)
        except Exception as e:
            import pickle

            pickle.dump(e, open("err", "wb"))

    def update(self, docs, upsert=False):
        """if id does not exist in the target_collection,
        the update will be ignored except if upsert is True
        """
        bulk = []
        for doc in docs:
            bulk.append(UpdateOne({"_id": doc["_id"]}, {"$set": doc}, upsert=upsert))

        if bulk:
            result = self.target_collection.bulk_write(bulk)
            # if doc is the same, it'll be matched but not modified.
            # but for us, it's been processed. if upserted, then it can't be matched
            # before (so matched count doesn't include upserted). finally, it's only update
            # ops, so don't count inserted_count and deleted_count
            return result.matched_count + result.upserted_count
        return 0

    def update_diff(self, diff, extra=None):
        """update a doc based on the diff returned from diff.diff_doc
        "extra" can be passed (as a dictionary) to add common fields to the
        updated doc, e.g. a timestamp.
        """
        extra = extra or {}
        _updates = {}
        _add_d = dict(list(diff.get("add", {}).items()) + list(diff.get("update", {}).items()))
        if _add_d or extra:
            if extra:
                _add_d.update(extra)
            _updates["$set"] = _add_d
        if diff.get("delete", None):
            # _updates['$unset'] = dict([(x, 1) for x in diff['delete']])    # TODO: remove this line, rewritten using dict comprehension
            _updates["$unset"] = {x: 1 for x in diff["delete"]}
        res = self.target_collection.update_one({"_id": diff["_id"]}, _updates, upsert=False)
        return res.modified_count

    def drop(self):
        self.target_collection.drop()

    def get_id_list(self):
        return [x["_id"] for x in self.target_collection.find(projection=[], manipulate=False)]

    def get_from_id(self, id):
        return self.target_collection.find_one({"_id": id})

    def mget_from_ids(self, ids, asiter=False):
        """ids is an id list.
        returned doc list should be in the same order of the
          input ids. non-existing ids are ignored.
        """
        # this does not return doc in the same order of ids
        cur = self.target_collection.find({"_id": {"$in": ids}})
        if asiter:
            return cur
        else:
            return [doc for doc in cur]

    def count_from_ids(self, ids, step=100000):
        """return the count of docs matching with input ids
        normally, it does not need to query in batches, but MongoDB
        has a BSON size limit of 16M bytes, so too many ids will raise a
        pymongo.errors.DocumentTooLarge error.
        """
        total_cnt = 0
        for i in range(0, len(ids), step):
            _ids = ids[i : i + step]
            _cnt = self.target_collection.count_documents({"_id": {"$in": _ids}})
            total_cnt += _cnt
        return total_cnt

    def finalize(self):
        """flush all pending writes."""
        # no need to flush, fsync is used for backups. also, this locks the whole
        # database for reads...
        # self.target_collection.database.client.fsync(async=True)

    def remove_from_ids(self, ids, step=10000):
        deleted = 0
        for i in range(0, len(ids), step):
            res = self.target_collection.delete_many({"_id": {"$in": ids[i : i + step]}})
            deleted += res.deleted_count
        return deleted


# backward-compatible
DocMongoDBBackend = DocMongoBackend


class DocESBackend(DocBackendBase):
    name = "es"

    def __init__(self, esidxer=None):
        """esidxer is an instance of utils.es.ESIndexer class."""
        if type(esidxer) == partial:
            self._target_esidxer_provider = esidxer
            self._target_esidxer = None
        else:
            self._target_esidxer_provider = None
            self._target_esidxer = esidxer

    @property
    def target_esidxer(self):
        if not self._target_esidxer:
            self._target_esidxer = self._target_esidxer_provider()
        return self._target_esidxer

    @property
    def target_name(self):
        return self.target_esidxer._index

    @property
    def target_alias(self):
        try:
            alias_info = self.target_esidxer.get_alias(self.target_name) or {}
            return list(alias_info[self.target_name]["aliases"].keys())[0]
        except Exception:
            return

    @property
    def version(self):
        try:
            mapping = self.target_esidxer.get_mapping_meta()
            if mapping.get("_meta"):
                return mapping["_meta"].get("build_version")
        except NotFoundError:
            # index doesn't even exist
            return None

    def prepare(self, update_mapping=True):
        self.target_esidxer.create_index()
        self.target_esidxer.verify_mapping(update_mapping=update_mapping)

    def count(self):
        try:
            return self.target_esidxer.count()
        except TransportError:
            return None

    def insert(self, doc_li):
        self.target_esidxer.add_docs(doc_li)

    def update(self, id, extra_doc):
        self.target_esidxer.update(id, extra_doc, bulk=True)

    def drop(self):
        conn = self.target_esidxer.conn
        index_name = self.target_esidxer.ES_INDEX_NAME
        index_type = self.target_esidxer.ES_INDEX_TYPE

        # Check if index_type exists
        try:
            conn.get_mapping(index_type, index_name)
        except NotFoundError:
            return
        return conn.delete_mapping(index_name, index_type)

    def finalize(self):
        conn = self.target_esidxer.conn
        conn.indices.flush()
        conn.indices.refresh()
        self.target_esidxer.optimize()

    def get_id_list(self, step=None):
        return self.target_esidxer.get_id_list(step=step)

    def get_from_id(self, id):
        return self.target_esidxer.get_biothing(id, only_source=True)

    def mget_from_ids(self, ids, step=100000, only_source=True, asiter=True, **kwargs):
        """ids is an id list. always return a generator"""
        return self.target_esidxer.get_docs(ids, step=step, only_source=only_source, **kwargs)

    def remove_from_ids(self, ids, step=10000):
        self.target_esidxer.delete_docs(ids, step=step)

    def query(self, query=None, verbose=False, step=10000, scroll="10m", only_source=True, **kwargs):
        """Function that takes a query and returns an iterator to query results."""
        try:
            return self.target_esidxer.doc_feeder(
                query=query, verbose=verbose, step=step, scroll=scroll, only_source=only_source, **kwargs
            )
        except Exception:
            pass

    @classmethod
    def create_from_options(cls, options):
        """Function that recreates itself from a DocBackendOptions class.  Probably a needless
        rewrite of __init__..."""
        if not options.es_index or not options.es_host or not options.es_doc_type:
            raise Exception(
                "Cannot create backend class from options, ensure that es_index, es_host, and es_doc_type are set"
            )
        return cls(ESIndexer(index=options.es_index, doc_type=options.es_doc_type, es_host=options.es_host))


class DocBackendOptions(object):
    def __init__(
        self, cls, es_index=None, es_host=None, es_doc_type=None, mongo_target_db=None, mongo_target_collection=None
    ):
        self.cls = cls
        self.es_index = es_index
        self.es_host = es_host
        self.es_doc_type = es_doc_type
        self.mongo_target_db = mongo_target_db
        self.mongo_target_collection = mongo_target_collection



"""
    Biothings API

    Support running biothings.web as a module

    >> python -m biothings.web
    >> python -m biothings.web --dir=~/mygene.info/src
    >> python -m biothings.web --dir=~/mygene.info/src --conf=config_web
    >> python -m biothings.web --conf=biothings.web.settings.default

    See more supported parameters in biothings.web.launcher.

"""

from biothings.web.launcher import main

main()



import pathlib
from typing import Callable, Generator, Iterable, Optional
from urllib.parse import parse_qsl, urlparse

import orjson


def ndjson_parser(
    patterns: Optional[Iterable[str]] = None,
) -> Callable[[str], Generator[dict, None, None]]:
    """
    Create NDJSON Parser given filename patterns

    For use with manifest.json based plugins.
    Caveat: Only handles valid NDJSON (no extra newlines, UTF8, etc.)

    Args:
        patterns: glob-compatible patterns for filenames, like *.ndjson, data*.ndjson

    Returns:
        parser_func: Generator that takes in a data_folder and returns documents from
            NDJSON files that matches the filename patterns
    """
    if patterns is None:
        raise TypeError("Must provide keyword argument patterns to" "match files for NDJSON Parser")

    def ndjson_parser_func(data_folder):
        work_dir = pathlib.Path(data_folder)
        for pattern in patterns:
            for filename in work_dir.glob(pattern):
                with open(filename, "rb") as f:
                    for line in f:
                        doc = orjson.loads(line)
                        yield doc

    return ndjson_parser_func


def json_array_parser(
    patterns: Optional[Iterable[str]] = None,
) -> Callable[[str], Generator[dict, None, None]]:
    """
    Create JSON Array Parser given filename patterns

    For use with manifest.json based plugins. The data comes in a JSON that is
    an JSON array, containing multiple documents.

    Args:
        patterns: glob-compatible patterns for filenames, like *.json, data*.json

    Returns:
        parser_func
    """
    if patterns is None:
        raise TypeError("Must provide keyword argument patterns to" "match files for JSON Array Parser")

    def json_array_parser(data_folder):
        work_dir = pathlib.Path(data_folder)
        for pattern in patterns:
            for filename in work_dir.glob(pattern):
                with open(filename, "r") as f:
                    data = orjson.loads(f.read())
                    try:
                        iterator = iter(data)
                    except TypeError:
                        raise RuntimeError(f"{filename} does not contain a valid" "JSON Array")
                    for doc in iterator:
                        yield doc

    return json_array_parser


def docker_source_info_parser(url):
    """
    :param url: file url include docker connection string
        format: docker://CONNECTION_NAME?image=DOCKER_IMAGE&tag=TAG&dump_command="python run.py"&path=/path/to/file
        the CONNECTION_NAME must be defined in the biothings Hub config.
        example:
        docker://CONNECTION_NAME?image=docker_image&tag=docker_tag&dump_command="python run.py"&path=/path/to/file
        docker://CONNECTION_NAME?image=docker_image&tag=docker_tag&dump_command="python run.py"&path=/path/to/file
        docker://CONNECTION_NAME?image=docker_image&tag=docker_tag&dump_command="python run.py"&path=/path/to/file
        docker"//CONNECTION_NAME?image=docker_image&tag=docker_tag&dump_command="python run.py"&path=/path/to/file
    :return:

    """
    parsed = urlparse(url)
    query = dict(parse_qsl(parsed.query))
    image = query.get("image")
    image_tag = query.get("tag")
    dump_command = query.get("dump_command")
    keep_container = query.get("keep_container")
    container_name = query.get("container_name")
    get_version_cmd = query.get("get_version_cmd")
    if keep_container:
        keep_container = keep_container.lower() in {"true", "yes", "1", "y"}
    if dump_command:
        dump_command = dump_command.strip('"')
    if get_version_cmd:
        get_version_cmd = get_version_cmd.strip('"')
    if not image_tag:
        image_tag = "latest"
    docker_image = image and f"{image}:{image_tag}" or None
    source_config = {
        "docker_image": docker_image,
        "path": query.get("path"),
        "dump_command": dump_command,
        "connection_name": parsed.netloc,
        "container_name": container_name,
        "keep_container": keep_container,
        "get_version_cmd": get_version_cmd,
    }
    if keep_container is None:
        # remove keep_container if not set, so that later we can check its value from image/container metadata
        source_config.pop("keep_container")
    return source_config



class WebAPIValidator:
    def validate(self, config):
        # Compatibility

        if getattr(config, "API_PREFIX", None) is not None:
            config.APP_PREFIX = config.API_PREFIX
        if getattr(config, "API_VERSION", None) is not None:
            config.APP_VERSION = config.API_VERSION

        assert config.APP_VERSION or config.APP_PREFIX, (
            "Require at least one of the follwing settings:"
            "(APP_VERSION, APP_PREFIX) to create a layer of"
            "separation for the default biothings routes."
        )


class DBParamValidator:
    def validate(self, config):
        assert isinstance(config.ES_INDICES, dict)

        # compatibility settings to convert ES_INDEX/ES_DOC_TYPE settings to ES_INDICES
        if getattr(config, "ES_INDEX", None):
            from biothings.web.settings import default

            if config.ES_INDICES is default.ES_INDICES:
                config.ES_INDICES = {}
            else:  # combine with the user provided value
                config.ES_INDICES = dict(config.ES_INDICES)
            # _doc_type can be None if not provided, in this case, ES_INDEX value will be
            # set to "None" key in ES_INDICES as the default index used in the handlers
            _doc_type = getattr(config, "ES_DOC_TYPE", None) or None
            config.ES_INDICES[_doc_type] = config.ES_INDEX

        ERROR = "UNSUPPORTED SETTINGS."
        # encountering the following attributes indicate
        # the application is built for a previous version.
        # to ensure delivering intended behaviors,
        # upgrade the config module of the application.
        assert not hasattr(config, "ES_SNIFF"), ERROR
        assert not hasattr(config, "ES_CLIENT_TIMEOUT"), ERROR


class SubmoduleValidator:
    def __init__(self):
        self._prefixes = set()

    def validate(self, config):
        assert config.APP_PREFIX, "APP_PREFIX must be defined for a submodule."
        assert config.APP_PREFIX not in self._prefixes, "APP_PREFIX conflicts."
        self._prefixes.add(config.APP_PREFIX)


class MongoParamValidaor:
    def validate(self, config):
        pass



import glob
import os
from subprocess import CalledProcessError, check_output

from setuptools import find_packages, setup

setup_path = os.path.dirname(__file__)


def read(fname):
    return open(os.path.join(setup_path, fname), encoding="utf8").read()


REPO_URL = "https://github.com/biothings/biothings.api"

# should fail if installed from source or from pypi,
# version gets set to MAJOR.MINOR.# commits on master branch if installed from pip repo
# otherwise to MAJOR.MINOR.MICRO as defined in biothings.version
try:
    command = ["git", "rev-list", "--count", "master"]
    NUM_COMMITS = check_output(command).strip().decode("utf-8")
except CalledProcessError:
    NUM_COMMITS = ""

# Calculate commit hash, should fail if installed from source or from pypi
try:
    command = ["git", "rev-parse", "HEAD"]
    COMMIT_HASH = check_output(command).strip().decode("utf-8")
except CalledProcessError:
    COMMIT_HASH = ""

# Write commit to file inside package, that can be read later
if COMMIT_HASH or NUM_COMMITS:
    with open("biothings/.git-info", "w", encoding="utf-8") as git_file:
        git_file.write(f"{REPO_URL}.git\n{COMMIT_HASH}\n{NUM_COMMITS}")


# very minimal requirement for running biothings.web
install_requires = [
    "requests>=2.21.0",
    'tornado==6.1.0; python_version < "3.7.0"',
    'tornado==6.2.0; python_version >= "3.7.0"',
    "gitpython>=3.1.0",
    "elasticsearch[async]>=7, <8",
    "elasticsearch-dsl>=7, <8",
    'singledispatchmethod; python_version < "3.8.0"',
    'dataclasses; python_version < "3.7.0"',
    "jmespath>=0.7.1,<2.0.0",  # support jmespath query parameter
    "PyYAML>=5.1",
    "orjson>=3.6.1",  # a faster json lib supports inf/nan and datetime, v3.6.1 is the last version supports Python 3.6
]

# extra requirements for biothings.web
web_extra_requires = [
    "msgpack>=0.6.1",  # support format=msgpack
    "sentry-sdk>=1.5.3",  # new sentry package
]

# extra requirements for biothings.web to use AWS OpenSearch
opensearch_requires = [
    "boto3",  # for AWS OpenSearch connection
    "requests-aws4auth",  # for AWS OpenSearch connection
]


# minimal requirements for running biothings.hub, e.g. in CLI mode
hubcore_requires = [
    "pymongo>=4.1.0,<5.0",  # support MongoDB 5.0 since v3.12.0
]

# extra requirements to run a full biothings.hub
hub_requires = hubcore_requires + [
    "beautifulsoup4",  # used in dumper.GoogleDriveDumper
    "aiocron==1.8",  # setup scheduled jobs
    "aiohttp==3.8.4",  # elasticsearch requires aiohttp>=3,<4
    "asyncssh==2.11.0",  # needs libffi-dev installed (apt-get)
    # "pymongo>=4.1.0,<5.0",  # support MongoDB 5.0 since v3.12.0
    "psutil",
    "jsonpointer",  # for utils.jsonpatch
    "IPython",  # for interactive hub console
    "multiprocessing_on_dill",  # can replace pickler in concurrent.futures
    "dill",  # a pickle alternative with extra object type support
    "prettytable",  # diff report renderer
    "sockjs-tornado==1.0.7",  # websocket server for HubServer
    "jsonschema>=2.6.0",
    "pip",  # auto-install requirements from plugins
    # 'pandas==1.0.1',    # json with inf/nan and more to come (replaced by orjson below now)
    # 'orjson>=3.5.2',    # this is a faster json lib support inf/nan and datetime
    # "yapf",  # code reformatter, better results than autopep8
    "black",  # code formatter
    "boto3",  # for aws S3 access
    "requests-aws4auth",  # aws s3 auth requests for autohub
    'networkx>=2.1,<3.0; python_version >= "3.7.0"',  # datatransform
    'networkx>=2.1,<2.6; python_version < "3.7.0"',  # datatransform
    "biothings_client>=0.2.6",  # datatransform (api client)
    "cryptography>=38.0.3",  # for generate ssh keys, ssl cert.
]

# minimal requirements for to run biothings CLI
cli_requires = [
    "typer[all]>=0.9.0",  # required for CLI, also install rich package with [all]
]

docker_requires = [
    "docker>=6.0.1",  # Docker SDK for Python, required for dockerdumper support
]

docker_ssh_requires = [
    "docker[ssh]>=6.0.1",  # install paramiko, only required when connecting docker using the ssh:// protocol
]
# extra requirements to develop biothings
dev_requires = [
    "pytest",
    "pytest-mock",
    "pre-commit==2.17.0",
]

# extra requirements for building docs
docs_requires = ["sphinx>=2.4.3", "sphinx_rtd_theme>=1.0.0", "sphinx_reredirects>=0.0.1"]


setup(
    name="biothings",
    version=__import__("biothings").__version__,
    author="The BioThings Team",
    author_email="cwu@scripps.edu",
    description="a toolkit for building high-performance data APIs in biology",
    license="Apache License, Version 2.0",
    keywords="biology annotation web service client api",
    url="https://biothings.io",
    project_urls={
        "Documentation": "https://docs.biothings.io",
        "Source": REPO_URL,
        "Tracker": f"{REPO_URL}/issues",
        "Changelog": f"{REPO_URL}/blob/master/CHANGES.txt",
        "Twitter": "https://twitter.com/biothingsapi",
    },
    packages=find_packages(exclude=["tests"]),
    package_data={"": ["*.html", "*.py.tpl"]},
    include_package_data=True,
    scripts=list(glob.glob("biothings/bin/*")),
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    install_requires=install_requires,
    extras_require={
        "web_extra": web_extra_requires,
        "opensearch": opensearch_requires,
        "hub": hub_requires + cli_requires,
        "cli": hubcore_requires + cli_requires,
        "docker": docker_requires,
        "docker_ssh": docker_ssh_requires,
        "dev": web_extra_requires + hub_requires + cli_requires + docker_requires + dev_requires + docs_requires,
    },
    entry_points={
        "console_scripts": ["biothings-cli=biothings.cli:main"],
    },
)



import hashlib
import logging
import os
import pickle
from functools import partial

import elasticsearch
import elasticsearch_dsl
import requests
from elasticsearch import AIOHttpConnection, RequestsHttpConnection as _Conn
from tornado.ioloop import IOLoop

from biothings.utils.common import run_once

try:
    import boto3
    from requests_aws4auth import AWS4Auth

    aws_avail = True
except ImportError:
    # only needed for connecting to AWS OpenSearch
    aws_avail = False

logger = logging.getLogger(__name__)

_should_log = run_once()


def _log_pkg():
    es_ver = elasticsearch.__version__
    es_dsl_ver = elasticsearch_dsl.__version__
    logger.info("Elasticsearch Package Version: %s", ".".join(map(str, es_ver)))
    logger.info("Elasticsearch DSL Package Version: %s", ".".join(map(str, es_dsl_ver)))


def _log_db(client, uri):
    logger.info(client)


def _log_es(client, hosts):
    _log_db(client, hosts)

    # only perform health check with the async client
    # so that it doesn't slow down program start time
    if isinstance(client, elasticsearch.AsyncElasticsearch):

        async def log_cluster(async_client):
            cluster = await async_client.info()
            # not specifying timeout in the function above because
            # there could be a number of es tasks scheduled before
            # this call and would take the cluster a while to respond

            if _should_log():
                _log_pkg()

            cluster_name = cluster["cluster_name"]
            version = cluster["version"]["number"]

            logger.info("%s: %s %s", hosts, cluster_name, version)

        IOLoop.current().add_callback(log_cluster, client)


# ------------------------
#   Low Level Functions
# ------------------------


class _AsyncConn(AIOHttpConnection):
    def __init__(self, *args, **kwargs):
        self.aws_auth = None
        _auth = kwargs.get("http_auth")
        if _auth and hasattr(_auth, "region") and isinstance(_auth, AWS4Auth):
            self.aws_auth = _auth
            kwargs["http_auth"] = None
        super().__init__(*args, **kwargs)

    async def perform_request(self, method, url, params=None, body=None, timeout=None, ignore=(), headers=None):
        req = requests.PreparedRequest()
        req.prepare(method, self.host + url, headers, None, body, params)
        self.aws_auth(req)  # sign the request
        headers.update(req.headers)
        return await super().perform_request(method, url, params, body, timeout, ignore, headers)


# https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instance-identity-documents.html
AWS_META_URL = "http://169.254.169.254/latest/dynamic/instance-identity/document"


def get_es_client(hosts=None, async_=False, **settings):
    """Enhanced ES client initialization.

    Additionally support these parameters:
        async_: use AsyncElasticserach instead of Elasticsearch.
        aws: setup request signing and provide reasonable ES settings
            to access AWS OpenSearch, by default assuming it is on HTTPS.
        sniff: provide resonable default settings to enable client-side
            LB to an ES cluster. this param itself is not an ES param.
    """

    if settings.pop("aws", False):
        if not aws_avail:
            raise ImportError('"boto3" and "requests_aws4auth" are required for AWS OpenSearch')
        # find region
        session = boto3.Session()
        region = session.region_name

        if not region:  # not in ~/.aws/config
            region = os.environ.get("AWS_REGION")
        if not region:  # not in environment variable
            try:  # assume same-region service access
                res = requests.get(AWS_META_URL)
                region = res.json()["region"]
            except Exception:  # not running in VPC
                region = "us-west-2"  # default

        # find credentials
        credentials = session.get_credentials()
        awsauth = AWS4Auth(refreshable_credentials=credentials, region=region, service="es")

        _cc = _AsyncConn if async_ else _Conn
        settings.update(http_auth=awsauth, connection_class=_cc)
        settings.setdefault("use_ssl", True)
        settings.setdefault("verify_certs", True)

    # not evaluated when 'aws' flag is set because
    # AWS OpenSearch is internally load-balanced
    # and does not support client-side sniffing.
    elif settings.pop("sniff", False):
        settings.setdefault("sniff_on_start", True)
        settings.setdefault("sniff_on_connection_fail", True)
        settings.setdefault("sniffer_timeout", 60)

    if async_:
        from elasticsearch import AsyncElasticsearch

        client = AsyncElasticsearch
    else:
        from elasticsearch import Elasticsearch

        client = Elasticsearch

    return client(hosts, **settings)


def get_sql_client(uri, **settings):
    from sqlalchemy import create_engine

    return create_engine(uri, **settings).connect()


def get_mongo_client(uri, **settings):
    from pymongo import MongoClient

    return MongoClient(uri, **settings).get_default_database()


def _not_implemented_client():
    raise NotImplementedError()


# ------------------------
#   High Level Utilities
# ------------------------


class _ClientPool:
    def __init__(self, client_factory, async_factory, callback=None):
        self._client_factory = client_factory
        self._clients = {}

        self._async_client_factory = async_factory
        self._async_clients = {}

        self.callback = callback or _log_db

    @staticmethod
    def hash(config):
        _config = pickle.dumps(config)
        _hash = hashlib.md5(_config)
        return _hash.hexdigest()

    def _get_client(self, repo, factory, uri, settings):
        hash = self.hash((uri, settings))
        if hash in repo:
            return repo[hash]
        repo[hash] = factory(uri, **settings)
        self.callback(repo[hash], uri)
        return repo[hash]

    def get_client(self, uri, **settings):
        return self._get_client(self._clients, self._client_factory, uri, settings)

    def get_async_client(self, uri, **settings):
        return self._get_client(self._async_clients, self._async_client_factory, uri, settings)


es = _ClientPool(get_es_client, partial(get_es_client, async_=True), _log_es)
sql = _ClientPool(get_sql_client, _not_implemented_client)
mongo = _ClientPool(get_mongo_client, _not_implemented_client)



import hashlib

# import smtplib
import uuid
from collections import UserDict
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from ipaddress import IPv4Address, IPv6Address, ip_address
from pprint import pformat
from random import randint
from typing import Union
from urllib.parse import urlencode


class Event(UserDict):
    # HTTP PageView
    # Fields under __request__:
    # user_agent, user_ip, host, path, referer

    def __getattr__(self, name):
        return self["__request__"][name] or ""

    def _cid_v1(self):
        # Author: Cyrus Afrasiabi
        # Reference: https://github.com/kra3/py-ga-mob
        # More at: pyga.utils.generate_hash

        hash_val = 1

        if self.user_agent:
            hash_val = 0
            for ordinal in map(ord, self.user_agent[::-1]):
                hash_val = ((hash_val << 6) & 0xFFFFFFF) + ordinal + (ordinal << 14)
                left_most_7 = hash_val & 0xFE00000
                if left_most_7 != 0:
                    hash_val ^= left_most_7 >> 21

        return (randint(0, 0x7FFFFFFF) ^ hash_val) & 0x7FFFFFFF

    def _cid_v2(self):
        # Author: Zhongchao Qian
        # More at: https://github.com/biothings/biothings.api/pull/124
        """
        Generates a unique user ID

        Using the remote IP and client user agent, to produce a somewhat
        unique identifier for users. A UUID version 4 conforming to RFC 4122
        is produced which is generally acceptable. It is not entirely random,
        but looks random enough to the outside. Using a hash func. the user
        info is completely anonymized.

        Args:
            remote_ip: Client IP address as a string, IPv4 or IPv6
            user_agent: User agent string of the client
        """

        try:
            rip: Union[IPv4Address, IPv6Address] = ip_address(self.user_ip)
            ip_packed = rip.packed
        except ValueError:  # in the weird case I don't get an IP
            ip_packed = randint(0, 0xFFFFFFFF).to_bytes(4, "big")  # nosec

        h = hashlib.blake2b(digest_size=16, salt=b"biothings")
        h.update(ip_packed)
        h.update(self.user_agent.encode("utf-8", errors="replace"))

        d = bytearray(h.digest())
        # truncating hash is not that bad, fixing some bits should be okay, too
        d[6] = 0x40 | (d[6] & 0x0F)  # set version
        d[8] = 0x80 | (d[8] & 0x3F)  # set variant
        u = str(uuid.UUID(bytes=bytes(d)))
        return u

    def _cid(self, version):
        if version == 1:
            return self._cid_v1()
        elif version == 2:
            return self._cid_v2()

        # this is a required GA field
        raise ValueError("CID Version.")

    def to_GA_payload(self, tracking_id, cid_version=1):
        # by default implements
        # a GA PageView hit-type

        # In the future, consider adding additional
        # keys as cutomized dimensions or metrics.

        payload = {
            "v": 1,  # protocol version
            "t": "pageview",
            "tid": tracking_id,
            "cid": self._cid(cid_version),
            "uip": self.user_ip,
            "dh": self.host,
            "dp": self.path,
        }

        # add document referer
        if isinstance(self.referer, str):
            if len(self.referer) <= 2048:  # GA Limit
                payload["dr"] = self.referer

        # add user_agent
        if self.user_agent:
            payload["ua"] = self.user_agent

        # this also escapes payload vals
        return [urlencode(payload)]

    def to_GA4_payload(self, measurement_id, cid_version=1):
        # Document about page_view event: https://support.google.com/analytics/answer/9964640#pageviews&zippy=%2Cin-this-article
        # GA4 does not support [Document path as UA](https://developers.google.com/analytics/devguides/collection/protocol/v1/parameters#dp)
        # so we use location instead of host and path

        # TODO consider to use request.uri as page_location
        payload = {
            "name": "page_view",
            "params": _clean(
                {
                    "page_location": f"{self.host}{self.path}",
                    "page_title": self.path.strip("/").replace("/", "-"),
                }
            ),
        }

        # add document referer
        if isinstance(self.referer, str):
            # Parameter values (including item parameter values) must be 100 character or fewer.
            if len(self.referer) <= 100:
                payload["params"]["page_referrer"] = self.referer

        # add user_agent
        if self.user_agent:
            payload["params"]["user_agent"] = self.user_agent[:100]

        return [payload]

    def __str__(self):  # to facilitate logging
        return f"{type(self).__name__}({pformat(self)})"


def _clean(dict):
    return {k: v for k, v in dict.items() if v}


class GAEvent(Event):
    # GA Event
    # {
    #   "category": "video",
    #   "action": "play",
    #   "label": "sample.mpg",
    #   "value": "60"
    # }

    def to_GA_payload(self, tracking_id, cid_version=1):
        payloads = super().to_GA_payload(tracking_id, cid_version)
        if self.get("category") and self.get("action"):
            payloads.append(
                urlencode(
                    _clean(
                        {
                            "v": 1,  # protocol version
                            "t": "event",
                            "tid": tracking_id,
                            "cid": self._cid(cid_version),
                            "ec": self["category"],
                            "ea": self["action"],
                            "el": self.get("label", ""),
                            "ev": self.get("value", ""),
                        }
                    )
                )
            )
        for event in self.get("__secondary__", []):
            event["__request__"] = self["__request__"]
            payloads.extend(event.to_GA_payload(tracking_id, cid_version)[1:])
            # ignore the first event (pageview)
            # which is already generated once
        return payloads

    def to_GA4_payload(self, measurement_id, cid_version=1):
        payloads = super().to_GA4_payload(measurement_id, cid_version)
        if self.get("category") and self.get("action"):
            payloads.append(
                {
                    # Following this article https://support.google.com/analytics/answer/11091026?hl=en&ref_topic=11091421#zippy=%2Cin-this-article
                    # <action> in the Universal Analytics property maps to <event_name> in the Google Analytics 4
                    # and 'event_category', 'event_label', and 'value' — along with their respective values — map to parameters with values.
                    "name": self["action"],
                    "params": _clean(
                        {
                            "event_category": self["category"],
                            "event_label": self.get("label", ""),
                            "value": self.get("value", ""),
                        }
                    ),
                }
            )
        for event in self.get("__secondary__", []):
            event["__request__"] = self["__request__"]
            payloads.extend(event.to_GA4_payload(measurement_id, cid_version)[1:])
            # ignore the first event (pageview)
            # which is already generated once
        return payloads


class Message(Event):
    """
    Logical document that can be sent through services.
    Processable fields: title, body, url, url_text, image, image_altext
    Optionally define default field values below.
    """

    DEFAULTS = {
        "title": "Notification Message",
        "url_text": "View Details",
        "image_altext": "<IMAGE>",
    }

    def __getattr__(self, attr):
        # virtual attributes
        if attr in ("title", "body", "url", "url_text", "image", "image_altext"):
            if attr in self:
                return self[attr]
            if attr in self.DEFAULTS:
                return self.DEFAULTS[attr]
            return ""
        raise AttributeError()

    def to_ADF(self):
        """
        Generate ADF for Atlassian Jira payload. Overwrite this to build differently.
        https://developer.atlassian.com/cloud/jira/platform/apis/document/playground/
        """
        adf = {"version": 1, "type": "doc", "content": []}
        if self.body:
            adf["content"].append({"type": "paragraph", "content": [{"type": "text", "text": self.body}]})
        if self.url:
            adf["content"].append(
                {
                    "type": "paragraph",
                    "content": [
                        {
                            "type": "text",
                            "text": self.url_text,
                            "marks": [{"type": "link", "attrs": {"href": self.url}}],
                        }
                    ],
                }
            )
        return adf

    def to_slack_payload(self):
        """
        Generate slack webhook notification payload.
        https://api.slack.com/messaging/composing/layouts
        """
        blocks = []
        if self.title:
            blocks.append(
                {
                    "type": "header",
                    "text": {
                        "type": "plain_text",
                        "text": ":sparkles: " + self.title,
                        "emoji": True,
                    },
                }
            )
            blocks.append({"type": "divider"})
        if self.body:
            body = {
                "type": "section",
                "text": {"type": "mrkdwn", "text": self.body},
            }
            if self.image:
                body["accessory"] = {
                    "type": "image",
                    "image_url": self.image,
                    "alt_text": self.image_altext,
                }
            blocks.append(body)
        if self.url:
            blocks.append(
                {
                    "type": "section",
                    "text": {"type": "mrkdwn", "text": f"*<{self.url}|{self.url_text}>*"},
                }
            )
        return {"attachments": [{"blocks": blocks}]}

    def to_jira_payload(self, profile):
        """
        Combine notification message with project profile to
        genereate jira issue tracking ticket request payload.
        """
        return {
            "fields": {
                "project": {"id": profile.project_id},
                "summary": self.title,
                "issuetype": {"id": profile.issuetype_id},
                "assignee": {"id": profile.assignee_id},
                "reporter": {"id": profile.reporter_id},
                "priority": {"id": "3"},
                "labels": [profile.label],
                "description": self.to_ADF(),
            }
        }

    def to_email_payload(self, sendfrom, sendto):
        """
        Build a MIMEMultipart message that can be sent as an email.
        https://docs.aws.amazon.com/ses/latest/DeveloperGuide/examples-send-using-smtp.html
        """
        # Create message container - the correct MIME type is multipart/alternative.
        msg = MIMEMultipart("alternative")
        msg["Subject"] = self.title
        # msg['From'] = email.utils.formataddr((SENDERNAME, SENDER))
        msg["From"] = sendfrom
        msg["To"] = sendto

        # Comment or delete the next line if you are not using a configuration set
        # msg.add_header('X-SES-CONFIGURATION-SET',CONFIGURATION_SET)

        # Record the MIME types of both parts - text/plain and text/html.
        part1 = MIMEText(self.body, "plain")
        part2 = MIMEText(
            f"""
        <!DOCTYPE html>
        <html>
            <head>
                <title>{self.title}</title>
            </head>
            <body>
                <p>{self.body}</p>
            </body>
        </html>""",
            "html",
        )

        # Attach parts into message container.
        # According to RFC 2046, the last part of a multipart message, in this case
        # the HTML message, is best and preferred.
        msg.attach(part1)
        msg.attach(part2)

        return msg



import datetime
import glob
import inspect
import io
import logging
import os
import time
from collections.abc import Iterable
from functools import partial, wraps

import bson
import dateutil.parser as date_parser
from pymongo import DESCENDING, MongoClient
from pymongo.client_session import ClientSession
from pymongo.collection import Collection as PymongoCollection
from pymongo.database import Database as PymongoDatabase
from pymongo.errors import AutoReconnect

from biothings.utils.backend import DocESBackend, DocMongoBackend
from biothings.utils.common import (
    # timesofar,
    dotdict,
    get_compressed_outfile,
    get_random_string,
    iter_n,
    open_compressed_file,
)
from biothings.utils.hub_db import IDatabase

# stub, until set to real config module
config = None


def handle_autoreconnect(cls_instance, func):
    """
    After upgrading the pymongo package from 3.12 to 4.x, the "AutoReconnect: connection pool paused" problem appears quite often.
    It is not clear that the problem happens with our codebase, maybe a pymongo's problem.

    This function is an attempt to handle the AutoReconnect exception, without modifying our codebase.
    When the exception is raised, we just wait for some time, then retry.
    If the error still happens after MAX_RETRY, it must be a connection-related problem.
    We should stop retrying and raise error.

    Ref: https://github.com/newgene/biothings.api/pull/40#issuecomment-1185334545
    """

    MAX_RETRY = 30
    SLEEP_TIME = 0.5  # seconds

    def inner(*args, **kwargs):
        retry = 0
        while retry < MAX_RETRY:
            try:
                return func(*args, **kwargs)
            except AutoReconnect:
                retry += 1
                time.sleep(SLEEP_TIME)

        raise MaxRetryAutoReconnectException()

    return inner


class HandleAutoReconnectMixin:
    """This mixin will decor any non-hidden method with handle_autoreconnect decorator"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for name, func in inspect.getmembers(self, inspect.ismethod):
            if name.startswith("_"):
                continue
            setattr(self, name, handle_autoreconnect(self, func))


class MaxRetryAutoReconnectException(AutoReconnect):
    """Raised when we reach maximum retry to connect to Mongo server"""


class DummyCollection(dotdict):
    def count(self):
        return None

    def drop(self):
        pass

    def __getitem__(self, what):
        # FIXME
        return DummyCollection()  # ???


class DummyDatabase(dotdict):
    def collection_names(self):
        return []

    def __getitem__(self, what):
        return DummyCollection()


class Collection(HandleAutoReconnectMixin, PymongoCollection):
    # https://pymongo.readthedocs.io/en/4.1.1/migrate-to-pymongo4.html

    def __bool__(self):
        return self is not None

    def insert(self, doc_or_docs, *args, **kwargs):
        if isinstance(doc_or_docs, Iterable):
            return self.insert_many(doc_or_docs)
        else:
            return self.insert_one(doc_or_docs)

    def update(self, spec, doc, *args, **kwargs):
        if kwargs.pop("multi", None):
            return self.update_many(spec, doc, *args, **kwargs)
        else:
            return self.update_one(spec, doc, *args, **kwargs)

    def remove(self, spec_or_id=None, **kwargs):
        if kwargs.pop("multi", None):
            return self.delete_many(spec_or_id, **kwargs)
        else:
            return self.delete_one(spec_or_id, **kwargs)

    def save(self, doc, *args, **kwargs):
        if "_id" in doc:
            kwargs["upsert"] = True
            self.replace_one({"_id": doc["_id"]}, doc, *args, **kwargs)
            return doc["_id"]
        else:
            res = self.insert_one(doc, *args, **kwargs)
            return res.inserted_id

    def count(self, _filter=None, **kwargs):
        if _filter:
            return self.count_documents(_filter, **kwargs)
        return self.estimated_document_count(**kwargs)


class Database(HandleAutoReconnectMixin, PymongoDatabase):
    def __bool__(self):
        return self is not None

    def __getitem__(self, name):
        return Collection(self, name)

    def collection_names(self, include_system_collections=True, session=None):
        _filter = None
        if not include_system_collections:
            _filter = {"name": {"$regex": r"^(?!system\\.)"}}
        return self.list_collection_names(session=session, filter=_filter)


class DatabaseClient(HandleAutoReconnectMixin, MongoClient, IDatabase):
    def __getitem__(self, name):
        return Database(self, name)


def requires_config(func):
    @wraps(func)
    def func_wrapper(*args, **kwargs):
        global config
        if not config:
            try:
                from biothings import config as config_mod

                config = config_mod
            except ImportError:
                raise Exception("call biothings.config_for_app() first")
        return func(*args, **kwargs)

    return func_wrapper


@requires_config
def get_conn(server, port):
    try:
        if config.DATA_SRC_SERVER_USERNAME and config.DATA_SRC_SERVER_PASSWORD:
            uri = f"mongodb://{config.DATA_SRC_SERVER_USERNAME}:{config.DATA_SRC_SERVER_PASSWORD}@{server}:{port}"
        else:
            uri = f"mongodb://{server}:{port}"
        conn = DatabaseClient(uri)
        return conn
    except (AttributeError, ValueError):
        # missing config variables (or invalid), we'll pretend it's a dummy connection to mongo
        # (dummy here means there really shouldn't be any call to get_conn() but mongo is too much tied to the code and needs more work to unlink it)
        return DummyDatabase()


@requires_config
def get_hub_db_conn():
    conn = DatabaseClient(config.HUB_DB_BACKEND["uri"])
    return conn


@requires_config
def get_src_conn():
    return get_conn(config.DATA_SRC_SERVER, getattr(config, "DATA_SRC_PORT", 27017))


@requires_config
def get_src_db(conn=None):
    conn = conn or get_src_conn()
    return conn[config.DATA_SRC_DATABASE]


@requires_config
def get_src_master(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.DATA_SRC_MASTER_COLLECTION]


@requires_config
def get_src_dump(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][getattr(config, "DATA_SRC_DUMP_COLLECTION", "src_dump")]


@requires_config
def get_src_build(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.DATA_SRC_BUILD_COLLECTION]


@requires_config
def get_src_build_config(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.DATA_SRC_BUILD_COLLECTION + "_config"]


@requires_config
def get_data_plugin(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.DATA_PLUGIN_COLLECTION]


@requires_config
def get_api(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.API_COLLECTION]


@requires_config
def get_cmd(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.CMD_COLLECTION]


@requires_config
def get_event(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][getattr(config, "EVENT_COLLECTION", "event")]


@requires_config
def get_hub_config(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][getattr(config, "HUB_CONFIG_COLLECTION", "hub_config")]


@requires_config
def get_last_command(conn=None):
    cmd = get_cmd(conn)
    cur = cmd.find({}, {"_id": 1}).sort("_id", DESCENDING).limit(1)
    return next(cur)


@requires_config
def get_target_conn():
    if config.DATA_TARGET_SERVER_USERNAME and config.DATA_TARGET_SERVER_PASSWORD:
        uri = "mongodb://{}:{}@{}:{}".format(
            config.DATA_TARGET_SERVER_USERNAME,
            config.DATA_TARGET_SERVER_PASSWORD,
            config.DATA_TARGET_SERVER,
            config.DATA_TARGET_PORT,
        )
    else:
        uri = "mongodb://{}:{}".format(config.DATA_TARGET_SERVER, config.DATA_TARGET_PORT)
    conn = DatabaseClient(uri)
    return conn


@requires_config
def get_target_db(conn=None):
    conn = conn or get_target_conn()
    return conn[config.DATA_TARGET_DATABASE]


@requires_config
def get_target_master(conn=None):
    conn = conn or get_target_conn()
    return conn[config.DATA_TARGET_DATABASE][config.DATA_TARGET_MASTER_COLLECTION]


@requires_config
def get_source_fullname(col_name):
    """
    Assuming col_name is a collection created from an upload process, find the main source & sub_source associated.
    """
    src_dump = get_src_dump()
    # "sources" in config is a list a collection names. src_dump _id is the name of the
    # resource but can have sub-resources with different collection names. We need
    # to query inner keys upload.job.*.step, which always contains the collection name
    info = src_dump.find_one(
        {
            "$where": 'function() {if(this.upload) {for(var index in this.upload.jobs) {if(this.upload.jobs[index].step == "%s") return this;}}}'
            % col_name
        }
    )
    if info:
        name = info["_id"]
        if name != col_name:
            # col_name was a sub-source name
            return "%s.%s" % (name, col_name)
        else:
            return name


def get_source_fullnames(col_names):
    main_sources = set()
    for col_name in col_names:
        main_source = get_source_fullname(col_name)
        if main_source:
            main_sources.add(main_source)
    return list(main_sources)


def doc_feeder(collection, step=1000, s=None, e=None, inbatch=False, query=None, batch_callback=None, fields=None, logger=logging, session_refresh_interval=5):
    """
    An iterator returning docs in a collection, with batch query.

    Additional filter query can be passed via `query`, e.g., `doc_feeder(collection, query={'taxid': {'$in': [9606, 10090, 10116]}})`
    `batch_callback` is a callback function as `fn(index, t)`, called after every batch.
    `fields` is an optional parameter to restrict the fields to return.

    `session_refresh_interval` is 5 minutes by default. We call `refreshSessions` command every 5 minutes to keep a session alive, otherwise the session
        and all cursors attached (explicitly or implicitly) to the session will time out after idling for 30 minutes, even if we have `no_cursor_timeout` set
        True for a cursor. See https://www.mongodb.com/docs/manual/reference/command/refreshSessions/ and
        https://www.mongodb.com/docs/manual/reference/method/cursor.noCursorTimeout/#session-idle-timeout-overrides-nocursortimeout
    """

    if isinstance(collection, DocMongoBackend):
        collection = collection.target_collection

    # Determine the partition of the collection we iterate over.
    # E.g. when s = 10000 and e = 15000, the collection partition is range(10000, 15000) and the partition size is 10000 - 15000 = 5000.
    # We will return those 5000 docs one by one, or in 5 batches if `inbatch = True` and `step = 1000`.
    n = collection.count_documents(query or {})
    s = s or 0  # start of the collection partition (inclusive)
    e = e or n  # end of the collection partition (exclusive)
    logger.debug("Retrieving documents from collection '%s'. start = %d, end = %d, total = %d.", collection.name, s, e, n)

    cursor_index = s  # the integer index in the collection that the cursor is pointing to
    # job_start_time = time.time()
    batch_start_time = time.time()

    try:
        # Explicitly create a session object for the cursor to attach
        session: ClientSession = collection.database.client.start_session()
        session_uuid = session.session_id["id"].as_uuid()
        logger.debug("Session '%s' started for collection '%s'.", session_uuid, collection.name)

        cur = collection.find(query, no_cursor_timeout=True, projection=fields, session=session)
        # logger.debug("Querying '%s' from collection '%s' in session '%s'.", query, collection.name, session_uuid)
        if s:
            cur.skip(s)
            # logger.debug("Skipped %d documents from collection '%s'.", s, collection.name)
        if e:
            cur.limit(e - s)  # specify the maximum number of documents the cursor will return
            # logger.debug("Limited the cursor to fetch only %d documents (%d ~ %d) from collection '%s'.", e - s, s, e, collection.name)
        cur.batch_size(step)  # specify the number of documents the cursor returns per batch (transparent to cursor iterators)

        if inbatch:  # which specifies this `doc_feeder` function to return docs in batch. Not related to `cursor.batch_size()`
            doc_batch = []

        session_last_refresh_time = time.time()
        for doc in cur:
            session_current_time = time.time()
            # session_current_time and session_last_refresh_time are in second, while session_refresh_interval is in minute
            session_should_refresh = (session_current_time - session_last_refresh_time) > session_refresh_interval * 60
            if session_should_refresh:
                cmd_resp = collection.database.command("refreshSessions", [session.session_id], session=session)
                logger.debug("Session '%s' refreshed, resp=%s", session_uuid, cmd_resp)
                session_last_refresh_time = session_current_time

            if inbatch:
                doc_batch.append(doc)
            else:
                yield doc

            cursor_index += 1

            if cursor_index % step == 0:  # batch is full
                if inbatch:
                    yield doc_batch
                    doc_batch = []

                # logger.debug("Done.[%.1f%%,%s]", cursor_index * 100. / n, timesofar(batch_start_time))
                # logger.debug("Processing %d-%d documents...", cursor_index + 1, min(cursor_index + step, e))
                if batch_callback:
                    batch_callback(cursor_index, time.time() - batch_start_time)
                if cursor_index < e:
                    batch_start_time = time.time()

        if inbatch and doc_batch:
            # Important: need to yield the last batch here
            yield doc_batch

        # logger.debug("Finished.[total time: %s]", timesofar(job_start_time))
    finally:
        cur.close()
        logger.debug("Session '%s' to be ended.", session_uuid)
        session.end_session()


def get_cache_filename(col_name):
    cache_folder = getattr(config, "CACHE_FOLDER", None)
    if not cache_folder:
        return  # we don't even use cache, forget it
    cache_format = getattr(config, "CACHE_FORMAT", None)
    cache_file = os.path.join(config.CACHE_FOLDER, col_name)
    cache_file = cache_format and (cache_file + ".%s" % cache_format) or cache_file
    return cache_file


def invalidate_cache(col_name, col_type="src"):
    if col_type == "src":
        src_dump = get_src_dump()
        if "." not in col_name:
            fullname = get_source_fullname(col_name)
        # FIXME so we are assuming that col_name must contain ".", otherwise the following assertion would fail
        assert fullname, "Can't resolve source '%s' (does it exist ?)" % col_name

        main, sub = fullname.split(".")
        doc = src_dump.find_one({"_id": main})
        assert doc, "No such source '%s'" % main
        assert doc.get("upload", {}).get("jobs", {}).get(sub), "No such sub-source '%s'" % sub
        # this will make the cache too old
        doc["upload"]["jobs"][sub]["started_at"] = datetime.datetime.now()
        src_dump.update_one({"_id": main}, {"$set": {"upload.jobs.%s.started_at" % sub: datetime.datetime.now()}})
    elif col_type == "target":
        # just delete the cache file
        cache_file = get_cache_filename(col_name)
        if cache_file:
            try:
                os.remove(cache_file)
            except FileNotFoundError:
                pass


# TODO: this func deals with different backend, should not be in bt.utils.mongo
# and doc_feeder should do the same as this function regarding backend support
@requires_config
def id_feeder(col, batch_size=1000, build_cache=True, logger=logging, force_use=False, force_build=False, validate_only=False):
    """
    Return an iterator for all _ids in collection "col".

    Search for a valid cache file if available, if not, return a doc_feeder for that collection.
    Valid cache is a cache file that is newer than the collection.

    "db" can be "target" or "src".
    "build_cache" True will build a cache file as _ids are fetched, if no cache file was found.
    "force_use" True will use any existing cache file and won't check whether it's valid of not.
    "force_build" True will build a new cache even if current one exists and is valid.
    "validate_only" will directly return [] if the cache is valid (convenient way to check if the cache is valid).
    """
    src_db = get_src_db()
    col_ts = None  # timestamp of the collection
    found_meta = True

    if isinstance(col, DocMongoBackend):
        col = col.target_collection

    try:
        if col.database.name == config.DATA_TARGET_DATABASE:
            info = src_db["src_build"].find_one({"_id": col.name})
            if not info:
                logger.warning("Can't find information for target collection '%s'", col.name)
            else:
                col_ts = info.get("_meta", {}).get("build_date")
                col_ts = col_ts and date_parser.parse(col_ts).timestamp()
        elif col.database.name == config.DATA_SRC_DATABASE:
            src_dump = get_src_dump()
            info = src_dump.find_one(
                {
                    "$where": 'function() {if(this.upload) {for(var index in this.upload.jobs) {if(this.upload.jobs[index].step == "%s") return this;}}}'
                    % col.name
                }
            )
            if not info:
                logger.warning("Can't find information for source collection '%s'", col.name)
            else:
                col_ts = info["upload"]["jobs"][col.name]["started_at"].timestamp()
        else:
            logger.warning("Can't find metadata for collection '%s' (not a target, not a source collection)", col)
            found_meta = False
            build_cache = False
    except KeyError:
        logger.warning("Couldn't find timestamp in database for '%s'", col.name)
    except Exception as e:
        logger.info("%s is not a mongo collection, _id cache won't be built (error: %s)", col, e)
        build_cache = False

    # try to find a cache file
    use_cache = False
    cache_file = None
    cache_format = getattr(config, "CACHE_FORMAT", None)
    if found_meta and getattr(config, "CACHE_FOLDER", None):
        cache_file = get_cache_filename(col.name)
        try:
            # size of empty file differs depending on compression
            empty_size = {None: 0, "xz": 32, "gzip": 25, "bz2": 14}
            if force_build:
                logger.warning("Force building cache file")
                use_cache = False
            # check size, delete if invalid
            elif os.path.getsize(cache_file) <= empty_size.get(cache_format, 32):
                logger.warning("Cache file exists but is empty, delete it")
                os.remove(cache_file)
            elif force_use:
                use_cache = True
                logger.info("Force using cache file")
            else:
                cache_ts = os.path.getmtime(cache_file)  # Get DLM (Date Last Modified) of the cache file in timestamp format
                if col_ts and cache_ts >= col_ts:
                    cache_dt = datetime.datetime.fromtimestamp(cache_ts).isoformat()
                    col_dt = datetime.datetime.fromtimestamp(col_ts).isoformat()
                    logger.debug("Cache is valid, cache_datetime:%s >= collection_datetime:%s", cache_dt, col_dt)
                    use_cache = True
                else:
                    logger.info("Cache is too old, discard it")
        except FileNotFoundError:
            pass

    if use_cache:
        logger.debug("Found valid cache file for '%s': %s", col.name, cache_file)
        if validate_only:
            logger.debug("Only validating cache, now return")
            yield []
            return

        with open_compressed_file(cache_file) as cache_in:
            if cache_format:
                io_cache = io.TextIOWrapper(cache_in)
            else:
                io_cache = cache_in
            for ids in iter_n(io_cache, batch_size):
                yield [_id.strip() for _id in ids if _id.strip()]
    else:
        logger.debug("No cache file found (or invalid) for '%s', use doc_feeder", col.name)
        cache_out = None
        cache_temp = None
        if getattr(config, "CACHE_FOLDER", None) and config.CACHE_FOLDER and build_cache:
            if not os.path.exists(config.CACHE_FOLDER):
                os.makedirs(config.CACHE_FOLDER)
            cache_temp = f"{cache_file}._tmp_"
            # clean aborted cache file generation
            for tmp_cache in glob.glob(os.path.join(config.CACHE_FOLDER, f"{cache_temp}*")):
                logger.info("Removing aborted cache file '%s'", tmp_cache)
                os.remove(tmp_cache)
            # use temp file and rename once done
            cache_temp = f"{cache_temp}{get_random_string()}"
            cache_out = get_compressed_outfile(cache_temp, compress=cache_format)
            logger.info("Building cache file '%s'", cache_temp)
        else:
            logger.info("Can't build cache, cache not allowed or no cache folder")
            build_cache = False

        if isinstance(col, PymongoCollection):
            doc_feeder_func = partial(doc_feeder, col, step=batch_size, inbatch=True, fields={"_id": 1}, logger=logger)
        elif isinstance(col, DocMongoBackend):
            doc_feeder_func = partial(doc_feeder, col.target_collection, step=batch_size, inbatch=True, fields={"_id": 1}, logger=logger)
        elif isinstance(col, DocESBackend):
            # get_id_list directly return the _id, wrap it to match other
            # doc_feeder_func returned vals. Also return a batch of id
            def wrap_id():
                ids = []
                for _id in col.get_id_list(step=batch_size):
                    ids.append({"_id": _id})
                    if len(ids) >= batch_size:
                        yield ids
                        ids = []
                if ids:
                    yield ids

            doc_feeder_func = partial(wrap_id)
        else:
            raise Exception("Unknown backend %s" % col)

        for doc_ids in doc_feeder_func():
            doc_ids = [str(_doc["_id"]) for _doc in doc_ids]
            if build_cache:
                str_out = "\n".join(doc_ids) + "\n"
                if cache_format:
                    # assuming binary format (b/compressed)
                    cache_out.write(str_out.encode())
                else:
                    cache_out.write(str_out)
            yield doc_ids

        if build_cache:
            cache_out.close()
            cache_final = os.path.splitext(cache_temp)[0]
            try:
                os.rename(cache_temp, cache_final)
            except OSError:
                logger.exception("Couldn't set final cache filename, building cache failed")


def check_document_size(doc):
    """
    Return True if doc isn't too large for mongo DB
    """
    return len(bson.BSON.encode(doc)) < 16777216  # 16*1024*1024


def get_previous_collection(new_id):
    """
    Given 'new_id', an _id from src_build, as the "new" collection, automatically select an "old" collection.
    By default, src_build's documents will be sorted according to their name (_id) and old collection is the one just before new_id.

    Note: because there can be more than one build config used, the actual build config name is first determined using new_id collection name,
      then the find().sort() is done on collections containing that build config name.
    """
    # TODO: this is not compatible with a generic hub_db backend
    # TODO: this should return a collection with status=success
    col = get_src_build()
    doc = col.find_one({"_id": new_id})
    assert doc, "No build document found for '%s'" % new_id
    assert "build_config" in doc, "No build configuration found for document '%s'" % new_id
    assert doc["build_config"]["name"] == doc["build_config"]["_id"]

    docs = (
        get_src_build()
        .find(
            {
                "$and": [
                    {"started_at": {"$lte": doc["started_at"]}},
                    {"build_config.name": doc["build_config"]["name"]},
                    {"archived": {"$exists": 0}},
                ]
            },
            {"_id": 1},
        )
        .sort([("started_at", -1)])
        .limit(2)
    )
    _ids = [d["_id"] for d in docs]
    assert len(_ids) == 2, "Expecting 2 collection _ids, got: %s" % _ids
    assert _ids[0] == new_id, "Can't find collection _id '%s'" % new_id
    return _ids[1]



import warnings

warnings.filterwarnings("default", category=DeprecationWarning, module=__name__)
warnings.warn(
    "This module is deprecated, please use `biothings.utils.storage` module", DeprecationWarning, stacklevel=2
)
from biothings.utils.storage import *  # NOQA #F403



# from __future__ import annotations  # for cyclic type hints
# Use forward-references for compatible with python3.6
# ref https://peps.python.org/pep-0484/#forward-references

import locale
from datetime import datetime

from dateutil.parser import parse as dtparse

from biothings.utils.dataload import update_dict_recur
from biothings.utils.hub_db import get_source_fullname
from biothings.utils.jsondiff import make as make_json_diff

locale.setlocale(locale.LC_ALL, "")


class ReleaseNoteSrcBuildReader:
    # TODO shall we use biothings.hub.dataindex.indexer._Build_doc here?

    def __init__(self, src_build_doc: dict):
        self.src_build_doc = src_build_doc

        # If `self` is a "hot" src_build doc reader, it can refer to a "cold" reader to access the cold build info.
        # This works like a two-node linked list.
        self.cold_src_build_reader: "ReleaseNoteSrcBuildReader" = None

    @property
    def build_id(self) -> str:
        return self.src_build_doc["_id"]

    @property
    def build_version(self) -> str:
        return self.src_build_doc.get("_meta", {}).get("build_version")

    @property
    def cold_collection_name(self) -> str:
        return self.src_build_doc.get("build_config", {}).get("cold_collection", None)

    def has_cold_collection(self) -> bool:
        return self.cold_collection_name is not None

    def attach_cold_src_build_reader(self, other: "ReleaseNoteSrcBuildReader"):
        """
        Attach a cold src_build reader.

        It's required that `self` is a hot src_builder reader and `other` is cold.
        """
        if not self.has_cold_collection():
            raise ValueError(
                f"{self.build_id} is not a hot src_build doc, "
                f"thus not able to attach a cold reader of {other.build_id}."
            )

        if other.has_cold_collection():
            raise ValueError(
                f"{other.build_id} is a hot src_build doc, "
                f"thus not able to be attached to the reader of {self.build_id}"
            )

        # src_build `_id`s and collection names are interchangeable
        # See https://github.com/biothings/biothings.api/blob/master/biothings/hub/databuild/builder.py#L311
        if self.cold_collection_name != other.build_id:
            raise ValueError(
                f"{self.build_id} has cold collection {self.cold_collection_name}, "
                f"while the reader to be attached is for {other.build_id}"
            )

        self.cold_src_build_reader = other

    @property
    def build_stats(self) -> dict:
        meta = self.src_build_doc.get("_meta", {})
        return meta.get("stats", {})

    def _get_datasource_stats(self) -> dict:
        return self.src_build_doc.get("merge_stats", {})

    def _get_datasource_versions(self) -> dict:
        meta = self.src_build_doc.get("_meta", {})

        # previous version format
        if "src_version" in meta:
            return meta["src_version"]

        # current version format
        src = meta.get("src", {})
        src_version = {src_name: src_info["version"] for src_name, src_info in src.items() if "version" in src_info}
        return src_version

    def _get_datasource_mapping(self) -> dict:
        return self.src_build_doc.get("mapping", {})

    @property
    def datasource_stats(self) -> dict:
        if not self.has_cold_collection():
            return self._get_datasource_stats()

        combined_stats = {
            **self._get_datasource_stats(),
            **self.cold_src_build_reader._get_datasource_stats(),
        }
        return combined_stats

    @property
    def datasource_versions(self) -> dict:
        if not self.has_cold_collection():
            return self._get_datasource_versions()

        combined_versions = {
            **self._get_datasource_versions(),
            **self.cold_src_build_reader._get_datasource_versions(),
        }
        return combined_versions

    @property
    def datasource_mapping(self) -> dict:
        if not self.has_cold_collection():
            return self._get_datasource_mapping()

        combined_mapping = {
            **self._get_datasource_mapping(),
            **self.cold_src_build_reader._get_datasource_mapping(),
        }
        return combined_mapping


class ReleaseNoteSrcBuildReaderAdapter:
    def __init__(self, src_build_reader: ReleaseNoteSrcBuildReader):
        self.src_build_reader = src_build_reader

    @classmethod
    def _get_datasource_fullname_stats(cls, datasource_stats) -> dict:
        """
        Receive a stat dictionary of <datasource_name>:<doc_count>, fetch the full datasource name, and
        return a new stat dictionary.

        If a full datasource name is two-tier (e.g. "gnomad.gnomad_exomes_hg19", as the full datasource name of
        "gnomad_exomes_hg19"), the returned dictionary is formed as:

            { <main_datasource_name> : { <sub_datasource_name> : { "_count" : <doc_count> } } }

        e.g.

            { "gnomad" : { "gnomad_exomes_hg19" : { "_count" : 12345678 } } }

        If a full datasource name is one-tier:
            CASE 1: the full datasource name is identical to the input datasource name,
                e.g. "cosmic" is the full name of "cosmic";
            CASE 2: the full datasource is None,
                e.g. when the input datasource name is "observed" or "total" in MyVariant, or "total_*" in MyGene
                In this case, the input datasource name is not a merge stat from a source but a custom field stat.

        the returned stats dictionary has the following structure:

            { <datasource_name> : { "_count" : <doc_count> } }

        e.g.

            { "cosmic" : { "_count" : 12345678 } }
            { "total" : { "_count" : 12345678 } }
        """
        result = {}
        for datasource_name, doc_count in datasource_stats.items():
            datasource_fullname = get_source_fullname(datasource_name)

            if (datasource_fullname is None) or (datasource_fullname == datasource_name):
                # one-tier fullname
                result[datasource_name] = {"_count": doc_count}
            else:
                # two-tier fullname
                main_name, sub_name = datasource_fullname.split(".")
                result.setdefault(main_name, {})
                result[main_name][sub_name] = {"_count": doc_count}

        return result

    @classmethod
    def _expand_datasource_versions(cls, datasource_versions) -> dict:
        """
        Receive a version dictionary of <datasource_name>:<version> (e.g. {"dbsnp" : "155"}), rewrite it to

            { <datasource_name> : {"_version" : <version>} }
        """
        return dict((k, {"_version": v}) for k, v in datasource_versions.items())

    @property
    def datasource_info(self):
        datasource_versions = self._expand_datasource_versions(self.src_build_reader.datasource_versions)
        datasource_stats = self._get_datasource_fullname_stats(self.src_build_reader.datasource_stats)

        datasource_info = update_dict_recur(datasource_versions, datasource_stats)
        return datasource_info

    @property
    def build_stats(self):
        # TODO this is the original logic, however I don't think it's necessary to apply get_source_fullname() to
        #  `build_stats.keys()`. E.g. in MyVariant, the `build_stats` keys are "total", "vcf", "hg19", and "observed",
        #  none of which has a two-tier full name. So the only effect of _get_datasource_fullname_stats() is to add a
        #  "_count" key to each of them.
        return self._get_datasource_fullname_stats(self.src_build_reader.build_stats)


class ReleaseNoteSource:
    def __init__(
        self,
        old_src_build_reader: ReleaseNoteSrcBuildReader,
        new_src_build_reader: ReleaseNoteSrcBuildReader,
        diff_stats_from_metadata_file: dict,
        addon_note: str,
    ):
        self.old_src_build_reader = old_src_build_reader
        self.new_src_build_reader = new_src_build_reader

        self.old_src_build_reader_adapter = ReleaseNoteSrcBuildReaderAdapter(self.old_src_build_reader)
        self.new_src_build_reader_adapter = ReleaseNoteSrcBuildReaderAdapter(self.new_src_build_reader)

        self.diff_stats_from_metadata_file = diff_stats_from_metadata_file
        self.addon_note = addon_note

    @classmethod
    def _make_stats_diff(cls, old: dict, new: dict):
        result = {
            "added": {},
            "deleted": {},
            "updated": {},
        }

        diff = make_json_diff(old, new)
        for item in diff:
            # get main source / main field
            key = item["path"].strip("/").split("/")[0]
            if item["op"] == "add":
                result["added"][key] = new[key]
            elif item["op"] == "remove":
                result["deleted"][key] = old[key]
            elif item["op"] == "replace":
                result["updated"][key] = {"new": new[key], "old": old[key]}
            else:
                raise ValueError("Unknown operation '%s' while computing changes" % item["op"])

        return result

    @classmethod
    def _make_mapping_diff(cls, old: dict, new: dict):
        def mapping_path_to_field_name(path: str) -> str:
            """
            Convert a JSON-Pointer path in a mapping json to a field name.

                E.g. path "/dbnsfp/properties/altai_neandertal" => field name "dbnsfp.altai_neandertal".

            Note that "properties" should not be included as part of a field name.
            The strategy here is iterate over the path components and remove any "properties" found at odd
            indices (1, 3, 5...).
            """
            path_components = path.strip("/").split("/")
            path_components = [
                path_components[i]
                for i in range(len(path_components))
                if (i % 2 == 0) or (i % 2 == 1 and path_components[i] != "properties")
            ]
            return ".".join(path_components)

        fields = {}

        diff = make_json_diff(old, new)
        for item in diff:
            if item["op"] in ("add", "remove", "replace"):
                field_name = mapping_path_to_field_name(item["path"])
                fields.setdefault(item["op"], []).append(field_name)
            elif item["op"] == "move":
                add_field_name = mapping_path_to_field_name(item["path"])
                remove_field_name = mapping_path_to_field_name(item["from"])

                fields.setdefault("add", []).append(add_field_name)
                fields.setdefault("remove", []).append(remove_field_name)
            else:
                raise ValueError("Unknown operation '%s' while computing changes" % item["op"])

        return fields

    def diff_build_stats(self) -> dict:
        # Read from the reader adapters, not the readers directly
        old_stats = self.old_src_build_reader_adapter.build_stats
        new_stats = self.new_src_build_reader_adapter.build_stats

        return self._make_stats_diff(old_stats, new_stats)

    def diff_datasource_info(self) -> dict:
        old_info = self.old_src_build_reader_adapter.datasource_info
        new_info = self.new_src_build_reader_adapter.datasource_info

        return self._make_stats_diff(old_info, new_info)

    def diff_datasource_mapping(self) -> dict:
        new_mapping = self.new_src_build_reader.datasource_mapping
        if not new_mapping:
            raise ValueError(f"New Mapping cannot be empty. Build id: {self.new_src_build_reader.build_id}")
        old_mapping = self.old_src_build_reader.datasource_mapping

        return self._make_mapping_diff(old_mapping, new_mapping)

    def to_dict(self) -> dict:
        result = {
            "old": {
                "_version": self.old_src_build_reader.build_version,
                "_count": self.old_src_build_reader.build_stats.get("total"),
            },
            "new": {
                "_version": self.new_src_build_reader.build_version,
                "_count": self.new_src_build_reader.build_stats.get("total"),
                "_fields": self.diff_datasource_mapping(),
                "_summary": self.diff_stats_from_metadata_file,
            },
            "stats": self.diff_build_stats(),
            "sources": self.diff_datasource_info(),
            "note": self.addon_note,
            "generated_on": str(datetime.now().astimezone()),
        }
        return result


class ReleaseNoteTxt(object):
    def __init__(self, source: ReleaseNoteSource):
        self.source = source  # member kept for debugging
        self.changes = source.to_dict()

    @classmethod
    def _format_number(cls, num, sign=None):
        try:
            sign_symbol = ""
            if sign:
                if num > 0:
                    sign_symbol = "+"
                elif num < 0:
                    sign_symbol = "-"

            num_str = locale.format_string("%d", abs(num), grouping=True)

            return "%s%s" % (sign_symbol, num_str)
        except TypeError:
            # something wrong with converting, maybe we don't even have a number to format...
            return "N.A"

    def save(self, filepath):
        try:
            import prettytable
        except ImportError:
            raise ImportError("Please install prettytable to use this rendered")

        txt = ""
        title = "Build version: '%s'" % self.changes["new"]["_version"]
        txt += title + "\n"
        txt += "".join(["="] * len(title)) + "\n"
        dt = dtparse(self.changes["generated_on"])
        txt += "Previous build version: '%s'\n" % self.changes["old"]["_version"]
        txt += "Generated on: %s\n" % dt.strftime("%Y-%m-%d at %H:%M:%S")
        txt += "\n"

        table = prettytable.PrettyTable(
            ["Updated datasource", "prev. release", "new release", "prev. # of docs", "new # of docs"]
        )
        table.align["Updated datasource"] = "l"
        table.align["prev. release"] = "c"
        table.align["new release"] = "c"
        table.align["prev. # of docs"] = "r"
        table.align["new # of docs"] = "r"

        for src, info in sorted(self.changes["sources"]["added"].items(), key=lambda e: e[0]):
            main_info = dict([(k, v) for k, v in info.items() if k.startswith("_")])
            sub_infos = dict([(k, v) for k, v in info.items() if not k.startswith("_")])
            if sub_infos:
                for sub, sub_info in sub_infos.items():
                    table.add_row(
                        ["%s.%s" % (src, sub), "-", main_info["_version"], "-", self._format_number(sub_info["_count"])]
                    )  # only _count avail there
            else:
                main_count = main_info.get("_count") and self._format_number(main_info["_count"]) or ""
                table.add_row([src, "-", main_info.get("_version", ""), "-", main_count])

        for src, info in sorted(self.changes["sources"]["deleted"].items(), key=lambda e: e[0]):
            main_info = dict([(k, v) for k, v in info.items() if k.startswith("_")])
            sub_infos = dict([(k, v) for k, v in info.items() if not k.startswith("_")])
            if sub_infos:
                for sub, sub_info in sub_infos.items():
                    table.add_row(
                        [
                            "%s.%s" % (src, sub),
                            main_info.get("_version", ""),
                            "-",
                            self._format_number(sub_info["_count"]),
                            "-",
                        ]
                    )  # only _count avail there
            else:
                main_count = main_info.get("_count") and self._format_number(main_info["_count"]) or ""
                table.add_row([src, main_info.get("_version", ""), "-", main_count, "-"])

        for src, info in sorted(self.changes["sources"]["updated"].items(), key=lambda e: e[0]):
            # extract information from main-source
            old_main_info = dict([(k, v) for k, v in info["old"].items() if k.startswith("_")])
            new_main_info = dict([(k, v) for k, v in info["new"].items() if k.startswith("_")])
            old_main_count = old_main_info.get("_count") and self._format_number(old_main_info["_count"]) or None
            new_main_count = new_main_info.get("_count") and self._format_number(new_main_info["_count"]) or None
            if old_main_count is None:
                assert new_main_count is None, (
                    "Sub-sources found for '%s', old and new count should " % src + "both be None. Info was: %s" % info
                )
                old_sub_infos = dict([(k, v) for k, v in info["old"].items() if not k.startswith("_")])
                new_sub_infos = dict([(k, v) for k, v in info["new"].items() if not k.startswith("_")])
                # old & new sub_infos should have the same structure (same existing keys)
                # so we just use one of them to explore
                if old_sub_infos:
                    assert new_sub_infos
                    for sub, sub_info in old_sub_infos.items():
                        table.add_row(
                            [
                                "%s.%s" % (src, sub),
                                old_main_info.get("_version", ""),
                                new_main_info.get("_version", ""),
                                self._format_number(sub_info["_count"]),
                                self._format_number(new_sub_infos[sub]["_count"]),
                            ]
                        )
            else:
                assert new_main_count is not None, (
                    "No sub-sources found, old and new count should NOT " + "both be None. Info was: %s" % info
                )
                table.add_row(
                    [
                        src,
                        old_main_info.get("_version", ""),
                        new_main_info.get("_version", ""),
                        old_main_count,
                        new_main_count,
                    ]
                )

        if table._rows:
            txt += table.get_string()
            txt += "\n"
        else:
            txt += "No datasource changed.\n"

        total_count = self.changes["new"].get("_count")
        if self.changes["sources"]["added"]:
            txt += "New datasource(s): %s\n" % ", ".join(sorted(list(self.changes["sources"]["added"])))
        if self.changes["sources"]["deleted"]:
            txt += "Deleted datasource(s): %s\n" % ", ".join(sorted(list(self.changes["sources"]["deleted"])))
        if self.changes["sources"]:
            txt += "\n"

        table = prettytable.PrettyTable(["Updated stats.", "previous", "new"])
        table.align["Updated stats."] = "l"
        table.align["previous"] = "r"
        table.align["new"] = "r"
        for stat_name, stat in sorted(self.changes["stats"]["added"].items(), key=lambda e: e[0]):
            table.add_row([stat_name, "-", self._format_number(stat["_count"])])
        for stat_name, stat in sorted(self.changes["stats"]["deleted"].items(), key=lambda e: e[0]):
            table.add_row([stat_name, self._format_number(stat["_count"]), "-"])
        for stat_name, stat in sorted(self.changes["stats"]["updated"].items(), key=lambda e: e[0]):
            table.add_row(
                [
                    stat_name,
                    self._format_number(stat["old"]["_count"]),
                    self._format_number(stat["new"]["_count"]),
                ]
            )

        if table._rows:
            txt += table.get_string()
            txt += "\n\n"

        if self.changes["new"]["_fields"]:
            new_fields = sorted(self.changes["new"]["_fields"].get("add", []))
            deleted_fields = sorted(self.changes["new"]["_fields"].get("remove", []))
            updated_fields = sorted(self.changes["new"]["_fields"].get("replace", []))
            if new_fields:
                txt += "New field(s): %s\n" % ", ".join(new_fields)
            if deleted_fields:
                txt += "Deleted field(s): %s\n" % ", ".join(deleted_fields)
            if updated_fields:
                txt += "Updated field(s): %s\n" % ", ".join(updated_fields)
            txt += "\n"

        if total_count is not None:
            txt += "Overall, %s documents in this release\n" % (self._format_number(total_count))

        if self.changes["new"]["_summary"]:
            sumups = []
            sumups.append("%s document(s) added" % self._format_number(self.changes["new"]["_summary"].get("add", 0)))
            sumups.append(
                "%s document(s) deleted" % self._format_number(self.changes["new"]["_summary"].get("delete", 0))
            )
            sumups.append(
                "%s document(s) updated" % self._format_number(self.changes["new"]["_summary"].get("update", 0))
            )
            txt += ", ".join(sumups) + "\n"
        else:
            txt += "No information available for added/deleted/updated documents\n"

        if self.changes.get("note"):
            txt += "\n"
            txt += "Note: %s\n" % self.changes["note"]

        with open(filepath, "w") as fout:
            fout.write(txt)

        return txt



"""
    Biothings Web Applications -

    define the routes and handlers a supported web framework would consume
    basing on a config file, typically named `config.py`, enhanced by
    :py:mod:`biothings.web.settings.configs`.

    The currently supported web frameworks are
    `Tornado <https://www.tornadoweb.org/en/stable/index.html>`_,
    `Flask <https://flask.palletsprojects.com/en/2.0.x/>`_, and
    `FastAPI <https://fastapi.tiangolo.com/>`_.

    The :py:mod:`biothings.web.launcher` can start the compatible HTTP servers
    basing on their interface. And the web applications delegate routes defined
    in the config file to handlers typically in :py:mod:`biothings.web.handlers`.

    +----------------+------------+--------------------------------+
    | Web Framework  | Interface  | Handlers                       |
    +================+============+================================+
    | Tornado        | Tornado    | biothings.web.handlers.*       |
    +----------------+------------+--------------------------------+
    | Flask          | WSGI       | biothings.web.handlers._flask  |
    +----------------+------------+--------------------------------+
    | FastAPI        | ASGI       | biothings.web.handlers._fastapi|
    +----------------+------------+--------------------------------+

"""

import inspect
import logging
from pprint import pformat
from pydoc import locate
from types import SimpleNamespace

import tornado.httpserver
import tornado.ioloop
import tornado.log
import tornado.web

from biothings.web.handlers import BaseAPIHandler, BaseQueryHandler
from biothings.web.services.namespace import BiothingsNamespace
from biothings.web.settings import configs

try:
    import sentry_sdk
    from sentry_sdk.integrations.tornado import TornadoIntegration
except ImportError:
    __SENTRY_INSTALLED__ = False
else:
    __SENTRY_INSTALLED__ = True

logger = logging.getLogger(__name__)


def load_class(kls):
    if inspect.isclass(kls):
        return kls
    if isinstance(kls, str):
        return locate(kls)
    raise ValueError()


class TornadoBiothingsAPI(tornado.web.Application):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.biothings = SimpleNamespace()

    @staticmethod
    def _get_settings(biothings, override=None):
        """
        Generates settings for tornado.web.Application. This result and the
        method below can define a tornado application to start a web server.
        """
        settings = {
            "biothings": biothings.config,
            "autoreload": False,
            "debug": False,
        }
        settings.update(override or {})
        supported_keywords = (
            "default_handler_class",
            "default_handler_args",
            "template_path",
            "log_function",
            "compress_response",
            "cookie_secret",
            "login_url",
            "static_path",
            "static_url_prefix",
        )

        for setting in supported_keywords:
            if hasattr(biothings.config, setting.upper()):
                if setting in settings:
                    logging.warning("Override config setting %s to %s.", setting, settings[setting])
                    continue
                settings[setting] = getattr(biothings.config, setting.upper())

        if __SENTRY_INSTALLED__ and biothings.config.SENTRY_CLIENT_KEY:
            sentry_sdk.init(
                dsn=biothings.config.SENTRY_CLIENT_KEY,
                # adjust this value to allow sentry to trace transactions:
                #   https://docs.sentry.io/platforms/python/guides/starlette/configuration/options/#traces-sample-rate
                traces_sample_rate=0.2,
                integrations=[TornadoIntegration()],
            )

        return settings

    @staticmethod
    def _get_handlers(biothings, addons=None):
        """
        Generates the tornado.web.Application `(regex, handler_class, options) tuples
        <http://www.tornadoweb.org/en/stable/web.html#application-configuration>`_.
        """
        handlers = {}
        addons = addons or []
        for rule in biothings.config.APP_LIST + addons:
            pattern = rule[0]
            handler = load_class(rule[1])
            setting = rule[2] if len(rule) == 3 else {}
            assert handler, rule[1]

            if "{typ}" in pattern or "{tps}" in pattern:
                if not issubclass(handler, BaseQueryHandler):
                    raise TypeError("Not a biothing_type-aware handler.")
                if "{tps}" in pattern and len(biothings.metadata.types) <= 1:
                    continue  # '{tps}' routes only valid for multi-type apps
                for biothing_type in biothings.metadata.types:
                    _pattern = pattern.format(
                        pre=biothings.config.APP_PREFIX,
                        ver=biothings.config.APP_VERSION,
                        typ=biothing_type,
                        tps=biothing_type,
                    ).replace("//", "/")
                    _setting = dict(setting)
                    _setting["biothing_type"] = biothing_type
                    handlers[_pattern] = (_pattern, handler, _setting)
            elif "{pre}" in pattern or "{ver}" in pattern:
                pattern = pattern.format(pre=biothings.config.APP_PREFIX, ver=biothings.config.APP_VERSION).replace(
                    "//", "/"
                )
                if "()" not in pattern:
                    handlers[pattern] = (pattern, handler, setting)
            else:  # no pattern translation
                handlers[pattern] = (pattern, handler, setting)

        handlers = list(handlers.values())
        logger.info("API Handlers:\n%s", pformat(handlers, width=200))
        return handlers

    @classmethod
    def get_app(cls, config, settings=None, handlers=None):
        """
        Return the tornado.web.Application defined by this config.
        **Additional** settings and handlers are accepted as parameters.
        """
        if isinstance(config, configs.ConfigModule):
            biothings = BiothingsNamespace(config)
            _handlers = BiothingsAPI._get_handlers(biothings, handlers)
            _settings = BiothingsAPI._get_settings(biothings, settings)
            app = cls(_handlers, **_settings)
            app.biothings = biothings
            app._populate_optionsets(config, _handlers)
            app._populate_handlers(_handlers)
            return app
        if isinstance(config, configs.ConfigPackage):
            biothings = BiothingsNamespace(config.root)
            _handlers = [(f"/{c.APP_PREFIX}/.*", cls.get_app(c, settings)) for c in config.modules]
            _settings = BiothingsAPI._get_settings(biothings, settings)
            app = cls(_handlers + handlers or [], **_settings)
            app.biothings = biothings
            # app._populate_optionsets(config, handlers or [])
            # app._populate_handlers(handlers or [])
            app._populate_optionsets(config, _handlers + handlers or [])
            app._populate_handlers(_handlers + handlers or [])
            return app
        raise TypeError()

    def _populate_optionsets(self, config, handlers):
        for handler in handlers:
            handler = handler[1]  # handler[0] is a matching pattern
            if inspect.isclass(handler) and issubclass(handler, BaseAPIHandler) and handler.name:
                handler_name = handler.name
                handler_options = handler.kwargs
                setting_attr = "_".join((handler_name, "kwargs")).upper()
                setting_options = getattr(config, setting_attr, {})
                self.biothings.optionsets.add(handler_name, setting_options)
                self.biothings.optionsets.add(handler_name, handler_options)

    def _populate_handlers(self, handlers):
        for handler in handlers:
            self.biothings.handlers[handler[0]] = handler[1]


try:
    from flask import Flask

    class FlaskBiothingsAPI(Flask):
        @classmethod
        def get_app(cls, config):
            app = cls(__name__)
            app.config["JSON_SORT_KEYS"] = False
            app.url_map.strict_slashes = False
            app.biothings = BiothingsNamespace(config)
            from biothings.web.handlers._flask import routes

            for route in routes:
                setting_attr = "_".join((route.name, "kwargs")).upper()
                setting_options = getattr(config, setting_attr, {})
                app.biothings.optionsets.add(route.name, setting_options)
                if isinstance(route.pattern, str):
                    route.pattern = [route.pattern]
                for pattern in route.pattern:
                    if "{typ}" in pattern:
                        assert len(app.biothings.metadata.types) == 1, (
                            "Currently Biothings API on Flask only " "supports single biothings_type configuration."
                        )
                    pattern = pattern.replace("{typ}", app.biothings.metadata.types[0])
                    pattern = pattern.replace("{ver}", app.biothings.config.APP_VERSION)
                    app.add_url_rule(pattern, route.name, route, methods=route.methods)
                    app.biothings.handlers[pattern] = route
            return app

except Exception as exc:  # noqa F841

    class FlaskBiothingsAPI:
        @classmethod
        def get_app(cls, config):
            raise exc


try:
    from fastapi import FastAPI
    from fastapi.middleware.wsgi import WSGIMiddleware

    class FastAPIBiothingsAPI(FastAPI):
        @classmethod
        def get_app(cls, config):
            app = cls()
            app.mount("/", WSGIMiddleware(FlaskBiothingsAPI.get_app(config)))
            return app

    # Native Implementation
    # -------------------------------------------------------------------------
    # class FastAPIBiothingsAPI(FastAPI):
    #     @classmethod
    #     def get_app(cls, config):
    #         from biothings.web.handlers import _fastapi
    #         _fastapi.biothings = BiothingsNamespace(config)
    #         app = cls()
    #         for route in _fastapi.routes:
    #             setting_attr = '_'.join((route.name, 'kwargs')).upper()
    #             setting_options = getattr(config, setting_attr, {})
    #             _fastapi.biothings.optionsets.add(route.name, setting_options)
    #             app.get(*route.args, **route.kwargs)(route)
    #         return app

except Exception as exc:  # noqa F841

    class FastAPIBiothingsAPI:
        @classmethod
        def get_app(cls, config):
            raise exc


BiothingsAPI = TornadoBiothingsAPI  # default



import asyncio
import logging
import math
import os
import pathlib
import shutil
import time
from pprint import pformat
from types import SimpleNamespace

import tornado.template
import typer
import yaml
from rich import box, print as rprint
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

import biothings.utils.inspect as btinspect
from biothings.utils import es
from biothings.utils.common import timesofar
from biothings.utils.dataload import dict_traverse
from biothings.utils.serializer import load_json, to_json
from biothings.utils.sqlite3 import get_src_db
from biothings.utils.workers import upload_worker


def get_logger(name=None):
    """Get a logger with the given name. If name is None, return the root logger."""
    # basicConfig has been setup in cli.py, so we don't need to do it again here
    # If everything works as expected, we can delete this block.
    # logging.basicConfig(
    #     level="INFO",
    #     format="%(message)s",
    #     datefmt="[%X]",
    #     handlers=[RichHandler(rich_tracebacks=True, show_path=False)],
    # )
    logger = logging.getLogger(name=None)
    return logger


def run_sync_or_async_job(func, *args, **kwargs):
    """When func is defined as either normal or async function/method, we will call this function properly and return the results.
    For an async function/method, we will use CLIJobManager to run it.
    """
    if asyncio.iscoroutinefunction(func):
        from biothings.utils.manager import CLIJobManager

        job_manager = CLIJobManager()
        kwargs["job_manager"] = job_manager
        return job_manager.loop.run_until_complete(func(*args, **kwargs))
    else:
        # otherwise just run it as normal
        return func(*args, **kwargs)


def load_plugin_managers(plugin_path, plugin_name=None, data_folder=None):
    """Load a data plugin from <plugin_path>, and return a tuple of (dumper_manager, upload_manager)"""
    from biothings import config as btconfig
    from biothings.hub.dataload.dumper import DumperManager
    from biothings.hub.dataload.uploader import UploaderManager
    from biothings.hub.dataplugin.assistant import LocalAssistant
    from biothings.hub.dataplugin.manager import DataPluginManager
    from biothings.utils.hub_db import get_data_plugin

    plugin_manager = DataPluginManager(job_manager=None)
    dmanager = DumperManager(job_manager=None)
    upload_manager = UploaderManager(job_manager=None)

    LocalAssistant.data_plugin_manager = plugin_manager
    LocalAssistant.dumper_manager = dmanager
    LocalAssistant.uploader_manager = upload_manager

    _plugin_path = pathlib.Path(plugin_path).resolve()
    btconfig.DATA_PLUGIN_FOLDER = _plugin_path.parent.as_posix()
    plugin_name = plugin_name or _plugin_path.name
    data_folder = data_folder or f"./{plugin_name}"
    assistant = LocalAssistant(f"local://{plugin_name}")
    # print(assistant.plugin_name, plugin_name, _plugin_path.as_posix(), btconfig.DATA_PLUGIN_FOLDER)
    dp = get_data_plugin()
    dp.remove({"_id": assistant.plugin_name})
    dp.insert_one(
        {
            "_id": assistant.plugin_name,
            "plugin": {
                "url": f"local://{plugin_name}",
                "type": assistant.plugin_type,
                "active": True,
            },
            "download": {
                "data_folder": data_folder,  # tmp path to your data plugin
            },
        }
    )
    p_loader = assistant.loader
    p_loader.load_plugin()

    return p_loader.__class__.dumper_manager, assistant.__class__.uploader_manager


def get_plugin_name(plugin_name=None, with_working_dir=True):
    """return a valid plugin name (the folder name contains a data plugin)
    When plugin_name is provided as None, it use the current working folder.
    when with_working_dir is True, returns (plugin_name, working_dir) tuple
    """
    working_dir = pathlib.Path().resolve()
    if plugin_name is None:
        plugin_name = working_dir.name
    else:
        valid_names = [f.name for f in os.scandir(working_dir) if f.is_dir() and not f.name.startswith(".")]
        if not plugin_name or plugin_name not in valid_names:
            rprint("[red]Please provide your data plugin name! [/red]")
            rprint("Choose from:\n    " + "\n    ".join(valid_names))
            raise typer.Exit(code=1)
    return plugin_name, working_dir if with_working_dir else plugin_name


def is_valid_data_plugin_dir(data_plugin_dir):
    """Return True/False if the given folder is a valid data plugin folder (contains either manifest.yaml or manifest.json)"""
    return (
        pathlib.Path(data_plugin_dir, "manifest.yaml").exists()
        or pathlib.Path(data_plugin_dir, "manifest.json").exists()
    )


def load_plugin(plugin_name=None, dumper=True, uploader=True, logger=None):
    """Return a plugin object for the given plugin_name.
    If dumper is True, include a dumper instance in the plugin object.
    If uploader is True, include uploader_classes in the plugin object.

    If <plugin_name> is not valid, raise the proper error and exit.
    """
    logger = logger or get_logger(__name__)
    _plugin_name, working_dir = get_plugin_name(plugin_name, with_working_dir=True)
    data_plugin_dir = pathlib.Path(working_dir) if plugin_name is None else pathlib.Path(working_dir, _plugin_name)
    if not is_valid_data_plugin_dir(data_plugin_dir):
        if plugin_name is None:
            err = (
                "This command must be run inside a data plugin folder. Please go to a data plugin folder and try again!"
            )
        else:
            err = f'The data plugin folder "{data_plugin_dir}" is not a valid data plugin folder. Please try another.'
        logger.error(err, extra={"markup": True})
        raise typer.Exit(1)

    if plugin_name is None:
        # current working_dir has the data plugin
        dumper_manager, uploader_manager = load_plugin_managers(working_dir, data_folder=".")
    else:
        dumper_manager, uploader_manager = load_plugin_managers(_plugin_name)

    current_plugin = SimpleNamespace(
        plugin_name=_plugin_name,
        data_plugin_dir=data_plugin_dir,
        in_plugin_dir=plugin_name is None,
    )
    if dumper:
        dumper_class = dumper_manager[_plugin_name][0]
        _dumper = dumper_class()
        _dumper.prepare()
        current_plugin.dumper = _dumper
    if uploader:
        uploader_classes = uploader_manager[_plugin_name]
        if not isinstance(uploader_classes, list):
            uploader_classes = [uploader_classes]
        current_plugin.uploader_classes = uploader_classes
    return current_plugin


########################
# for create command   #
########################


def do_create(name, multi_uploaders=False, parallelizer=False, logger=None):
    """Create a new data plugin from the template"""
    logger = logger or get_logger(__name__)
    working_dir = pathlib.Path().resolve()
    biothing_source_dir = pathlib.Path(__file__).parent.parent.resolve()
    template_dir = os.path.join(biothing_source_dir, "hub", "dataplugin", "templates")
    plugin_dir = os.path.join(working_dir, name)
    if os.path.isdir(plugin_dir):
        logger.error("Data plugin with the same name is already exists, please remove it before create")
        return exit(1)
    shutil.copytree(template_dir, plugin_dir)
    # create manifest file
    loader = tornado.template.Loader(plugin_dir)
    parsed_template = (
        loader.load("manifest.yaml.tpl").generate(multi_uploaders=multi_uploaders, parallelizer=parallelizer).decode()
    )
    manifest_file_path = os.path.join(working_dir, name, "manifest.yaml")
    with open(manifest_file_path, "w") as fh:
        fh.write(parsed_template)

    # remove manifest template
    os.unlink(f"{plugin_dir}/manifest.yaml.tpl")
    if not parallelizer:
        os.unlink(f"{plugin_dir}/parallelizer.py")
    logger.info(f"Successfully created data plugin template at: \n {plugin_dir}")


###############################
# for dump & upload command   #
###############################


def do_dump(plugin_name=None, show_dumped=True, logger=None):
    """Perform dump for the given plugin"""
    from biothings import config
    from biothings.utils import hub_db

    hub_db.setup(config)
    logger = logger or get_logger(__name__)
    _plugin = load_plugin(plugin_name, dumper=True, uploader=False, logger=logger)
    dumper = _plugin.dumper
    dumper.prepare()
    run_sync_or_async_job(dumper.create_todump_list, force=True)
    for item in dumper.to_dump:
        logger.info('Downloading remote data from "%s"...', item["remote"])
        dumper.download(item["remote"], item["local"])
        logger.info('Downloaded locally as "%s"', item["local"])
    dumper.steps = ["post"]
    dumper.post_dump()
    dumper.register_status("success")
    dumper.release_client()
    # cleanup
    # Commented out this line below. we should keep the dump info in src_dump collection for other cmds, e.g. upload, list etc
    # dumper.src_dump.remove({"_id": dumper.src_name})
    dp = hub_db.get_data_plugin()
    dp.remove({"_id": _plugin.plugin_name})
    data_folder = dumper.new_data_folder
    if show_dumped:
        logger.info("[green]Success![/green] :rocket:", extra={"markup": True})
        show_dumped_files(data_folder, _plugin.plugin_name)
    return _plugin


def do_upload(plugin_name=None, show_uploaded=True, logger=None):
    """Perform upload for the given list of uploader_classes"""

    logger = logger or get_logger(__name__)

    _plugin = load_plugin(plugin_name, dumper=False, uploader=True, logger=logger)
    for uploader_cls in _plugin.uploader_classes:
        uploader = uploader_cls.create(db_conn_info="")
        uploader.make_temp_collection()
        uploader.prepare()
        if not uploader.data_folder or not pathlib.Path(uploader.data_folder).exists():
            logger.error(
                'Data folder "%s" for "%s" is empty or does not exist yet. Have you run "dump" yet?',
                uploader.data_folder,
                uploader.fullname,
            )
            raise typer.Exit(1)
        else:
            upload_worker(
                uploader.fullname,
                uploader.__class__.storage_class,
                uploader.load_data,
                uploader.temp_collection_name,
                10000,
                1,
                uploader.data_folder,
                db=uploader.db,
            )
            uploader.switch_collection()
            uploader.keep_archive = 3  # keep 3 archived collections, that's probably good enough for CLI, default is 10
            uploader.clean_archived_collections()
    if show_uploaded:
        logger.info("[green]Success![/green] :rocket:", extra={"markup": True})
        show_uploaded_sources(pathlib.Path(_plugin.data_plugin_dir), _plugin.plugin_name)
    return _plugin


def do_dump_and_upload(plugin_name, logger=None):
    """Perform both dump and upload for the given plugin"""
    logger = logger or get_logger(__name__)
    _plugin = do_dump(plugin_name, show_dumped=False, logger=logger)
    do_upload(plugin_name, show_uploaded=False, logger=logger)
    logger.info("[green]Success![/green] :rocket:", extra={"markup": True})
    show_dumped_files(_plugin.dumper.new_data_folder, _plugin.plugin_name)
    show_uploaded_sources(pathlib.Path(_plugin.data_plugin_dir), _plugin.plugin_name)


########################
# for inspect command  #
########################


def process_inspect(source_name, mode, limit, merge, logger, do_validate, output=None):
    """Perform inspect for the given source. It's used in do_inspect function below"""
    VALID_INSPECT_MODES = ["jsonschema", "type", "mapping", "stats"]
    mode = mode.split(",")
    if "jsonschema" in mode:
        mode = ["jsonschema", "type"]
    for m in mode:
        if m not in VALID_INSPECT_MODES:
            logger.error('"%s" is not a valid inspect mode', m)
            raise typer.Exit(1)
    if not limit:
        limit = None
    sample = None
    clean = True

    t0 = time.time()
    data_provider = ("src", source_name)

    src_db = get_src_db()
    pre_mapping = "mapping" in mode
    src_cols = src_db[source_name]
    inspected = {}
    converters, mode = btinspect.get_converters(mode)
    for m in mode:
        inspected.setdefault(m, {})
    cur = src_cols.findv2()
    res = btinspect.inspect_docs(
        cur,
        mode=mode,
        clean=clean,
        merge=merge,
        logger=logger,
        pre_mapping=pre_mapping,
        limit=limit,
        sample=sample,
        metadata=False,
        auto_convert=False,
    )

    for m in mode:
        inspected[m] = btinspect.merge_record(inspected[m], res[m], m)
    for m in mode:
        if m == "mapping":
            try:
                inspected["mapping"] = es.generate_es_mapping(inspected["mapping"])
                # metadata for mapping only once generated
                inspected = btinspect.compute_metadata(inspected, m)
            except es.MappingError as e:
                inspected["mapping"] = {"pre-mapping": inspected["mapping"], "errors": e.args[1]}
        else:
            inspected = btinspect.compute_metadata(inspected, m)
    btinspect.run_converters(inspected, converters)

    res = btinspect.stringify_inspect_doc(inspected)
    _map = {"results": res, "data_provider": repr(data_provider), "duration": timesofar(t0)}

    # _map["started_at"] = started_at

    def clean_big_nums(k, v):
        # TODO: same with float/double? seems mongo handles more there ?
        if isinstance(v, int) and v > 2**64:
            return k, math.nan
        else:
            return k, v

    dict_traverse(_map, clean_big_nums)
    mapping = _map["results"].get("mapping", {})
    type_and_stats = {
        source_name: {
            _mode: btinspect.flatten_and_validate(_map["results"].get(_mode, {}), do_validate)
            for _mode in ["type", "stats"]
        }
    }
    mapping_table = None
    if "mapping" in mode and mapping:
        if mapping.get("errors"):
            mapping_table = f"[red]{mapping.get('errors')}[/red]"
        else:
            mapping_table = Table(
                title="[bold green]MAPPING[/bold green]",
                box=box.SQUARE_DOUBLE_HEAD,
                expand=False,
                show_lines=True,
            )
            mapping_table.add_column(f"Sub source name: [bold]{source_name}[/bold]", justify="left", style="cyan")
            mapping_table.add_row(to_json(mapping, indent=True, sort_keys=True))
    report = []
    problem_summary = []
    if "stats" in mode:
        report = type_and_stats[source_name]["stats"]
    elif "type" in mode:
        for item in type_and_stats[source_name]["type"]:
            item.pop("stats", None)
            report.append(item)
    console = Console()
    if report:
        panel = Panel(
            f"This is the field type and stats for datasource: [bold green]{source_name}[/bold green]\n"
            f"It provides a summary of the data structure, including: a map of all types involved in the data;\n"
            f"basic statistics, showing how volumetry fits over data structure.\n"
            f"The basic statistics include these fields:\n"
            f"* [italic]_count[/italic]: Total records\n"
            f"* [italic]_max[/italic]: Maximum value\n"
            f"* [italic]_min[/italic]: Minimum value\n"
            f"* [italic]_none[/italic]: number of records have no value",
            title=f"[bold green]{source_name}[/bold green]",
            box=box.HORIZONTALS,
        )
        for field in report:
            warnings = field.pop("warnings", [])
            field["warning"] = ""
            for warning in warnings:
                field_name = field.get("field_name")
                if field_name == "__root__":
                    continue
                warning_mgs = f"* [red]{warning['code']}[/red]: {warning['message']}"
                problem_summary.append(warning_mgs)
            if warnings:
                field["warning"] = ",".join(item["code"] for item in warnings)

        table = Table(
            title="[bold green]TYPE & STATS[/bold green]",
            box=box.SQUARE_DOUBLE_HEAD,
            show_lines=True,
        )
        table.add_column("Field name", justify="left", style="cyan")
        table.add_column("Field type", justify="center", style="magenta")
        table.add_column("Stats._count", justify="right", style="green")
        table.add_column("Stats._max", justify="right", style="green")
        table.add_column("Stats._min", justify="right", style="green")
        table.add_column("Stats._none", justify="right", style="green")
        table.add_column("Warning", justify="right", style="red")
        for item in report:
            if item["field_name"] == "__root__":
                continue
            field_name = f'[bold]{item["field_name"]}[/bold]'
            if item["warning"] != "":
                field_name += "[red] \u26a0[/red]" * len(item["warning"].split(","))
            table.add_row(
                field_name,
                str(item["field_type"]),
                str(item.get("stats", {}).get("_count", "")),
                str(item.get("stats", {}).get("_max", "")),
                str(item.get("stats", {}).get("_min", "")),
                str(item.get("stats", {}).get("_none", "")),
                item["warning"],
            )
        problem_panel = None
        if problem_summary:
            problem_panel = Panel("[yellow]Warnings:[/yellow]\n" + "\n".join(problem_summary))
        console.print(panel)
        if mapping_table:
            console.print(mapping_table)
        console.print(table)
        if problem_panel:
            console.print(problem_panel)
    elif mapping_table:
        console.print(mapping_table)

    if "mapping" in mode and mapping and output:
        # TODO: the following block is commented out because we don't need to append the mapping info to the existing. Delete this block if we verify it's not needed.
        # with open(output, "w+") as fp:
        #     current_content = fp.read()
        #     if current_content:
        #         current_content = load_json(current_content)
        #     else:
        #         current_content = {}
        #     current_content.update(mapping)
        #     fp.write(to_json(current_content, indent=True, sort_keys=True))
        with open(output, "w") as fp:
            fp.write(to_json(mapping, indent=True, sort_keys=True))
            rprint(f"[green]Successfully wrote the mapping info to the JSON file: [bold]{output}[/bold][/green]")


def do_inspect(
    plugin_name=None, sub_source_name=None, mode="type,stats", limit=None, merge=False, output=None, logger=None
):
    """Perform inspection on a data plugin."""
    logger = logger or get_logger(__name__)
    if not limit:
        limit = None

    _plugin = load_plugin(plugin_name, logger=logger)
    # source_full_name = _plugin.plugin_name if sub_source_name else f"{_plugin.plugin_name}.{sub_source_name}"
    if len(_plugin.uploader_classes) > 1:
        if not sub_source_name:
            logger.error('This is a multiple uploaders data plugin, so "--sub-source-name" must be provided!')
            logger.error(
                'Accepted values of "--sub-source-name" are: %s',
                ", ".join(uploader.name for uploader in _plugin.uploader_classes),
            )
            raise typer.Exit(code=1)
        logger.info(
            'Inspecting data plugin "%s" (sub_source_name="%s", mode="%s", limit=%s)',
            _plugin.plugin_name,
            sub_source_name,
            mode,
            limit,
        )
    else:
        logger.info('Inspecting data plugin "%s" (mode="%s", limit=%s)', _plugin.plugin_name, mode, limit)
    # table_space = get_uploaders(pathlib.Path(f"{working_dir}/{plugin_name}"))
    table_space = [item.name for item in _plugin.uploader_classes]
    if sub_source_name and sub_source_name not in table_space:
        logger.error('Your source name "%s" does not exits', sub_source_name)
        raise typer.Exit(code=1)
    if sub_source_name:
        process_inspect(sub_source_name, mode, limit, merge, logger=logger, do_validate=True, output=output)
    else:
        for source_name in table_space:
            process_inspect(source_name, mode, limit, merge, logger=logger, do_validate=True, output=output)


def get_manifest_content(working_dir):
    """return the manifest content of the data plugin in the working directory"""
    manifest_file_yml = os.path.join(working_dir, "manifest.yaml")
    manifest_file_json = os.path.join(working_dir, "manifest.json")
    if os.path.isfile(manifest_file_yml):
        manifest = yaml.safe_load(open(manifest_file_yml))
        return manifest
    elif os.path.isfile(manifest_file_json):
        manifest = load_json(open(manifest_file_json).read())
        return manifest
    else:
        raise FileNotFoundError("manifest file does not exits in current working directory")


########################
# for list command     #
########################


def get_uploaders(working_dir: pathlib.Path):
    """A helper function to get the uploaders from the manifest file in the working directory, used in show_uploaded_sources function below"""
    data_plugin_name = working_dir.name
    manifest = get_manifest_content(working_dir)
    upload_section = manifest.get("uploader")
    table_space = [data_plugin_name]
    if not upload_section:
        upload_sections = manifest.get("uploaders")
        table_space = [item["name"] for item in upload_sections]
    return table_space


def show_dumped_files(data_folder, plugin_name):
    """A helper function to show the dumped files in the data folder"""
    console = Console()
    if not os.path.isdir(data_folder) or not os.listdir(data_folder):
        console.print(
            Panel(
                f"[green]Source:[/green][bold] {plugin_name}[/bold]\n"
                + f"[green]Data Folder:[/green][bold] {data_folder}:[/bold]\n"
                + "Empty file!",
                title="[bold]Dump[/bold]",
                title_align="left",
            )
        )
    else:
        console.print(
            Panel(
                f"[green]Source:[/green][bold] {plugin_name}[/bold]\n"
                + f"[green]Data Folder:[/green][bold] {data_folder}:[/bold]\n    - "
                + "\n    - ".join(os.listdir(data_folder)),
                title="[bold]Dump[/bold]",
                title_align="left",
            )
        )


def get_uploaded_collections(src_db, uploaders):
    """A helper function to get the uploaded collections in the source database"""
    uploaded_sources = []
    archived_sources = []
    temp_sources = []
    for item in src_db.collection_names():
        if item in uploaders:
            uploaded_sources.append(item)
        for uploader_name in uploaders:
            if item.startswith(f"{uploader_name}_archive_"):
                archived_sources.append(item)
            if item.startswith(f"{uploader_name}_temp_"):
                temp_sources.append(item)
    return uploaded_sources, archived_sources, temp_sources


def show_uploaded_sources(working_dir, plugin_name):
    """A helper function to show the uploaded sources from given plugin."""
    console = Console()
    uploaders = get_uploaders(working_dir)
    src_db = get_src_db()
    uploaded_sources, archived_sources, temp_sources = get_uploaded_collections(src_db, uploaders)
    if not uploaded_sources:
        console.print(
            Panel(
                f"[green]Source:[/green] [bold]{plugin_name}[/bold]\n"
                + f"[green]DB path:[/green] [bold]{working_dir}/{src_db.dbfile}[/bold]\n"
                + f"[green]- Database:[/green] [bold]{src_db.name}[/bold]\n"
                + "Empty source!",
                title="[bold]Upload[/bold]",
                title_align="left",
            )
        )
    else:
        console.print(
            Panel(
                f"[green]Source:[/green] [bold]{plugin_name}[/bold]\n"
                + f"[green]DB path:[/green] [bold]{working_dir}/{src_db.dbfile}[/bold]\n"
                + f"[green]- Database:[/green] [bold]{src_db.name}[/bold]\n"
                + "    -[green] Collections:[/green] [bold]\n        "
                + "\n        ".join(uploaded_sources)
                + "[/bold] \n"
                + "    -[green] Archived collections:[/green][bold]\n        "
                + "\n        ".join(archived_sources)
                + "[/bold] \n"
                + "    -[green] Temporary collections:[/green][bold]\n        "
                + "\n        ".join(temp_sources),
                title="[bold]Upload[/bold]",
                title_align="left",
            )
        )


def show_hubdb_content():
    """Output hubdb content in a pretty format."""
    from biothings import config
    from biothings.utils import hub_db

    console = Console()
    hub_db.setup(config)
    coll_list = [
        hub_db.get_data_plugin(),
        hub_db.get_src_dump(),
        hub_db.get_src_master(),
        hub_db.get_hub_config(),
        hub_db.get_event(),
    ]
    hub_db_content = []

    for collection in coll_list:
        content = collection.find()
        if content:
            hub_db_content.append(f"[green]Collection:[/green] [bold]{collection.name}[/bold]\n{pformat(content)}")
    hub_db_content = "\n".join(hub_db_content)
    console.print(
        Panel(
            hub_db_content,
            title="[bold]Hubdb[/bold]",
            title_align="left",
        )
    )


def do_list(plugin_name=None, dump=False, upload=False, hubdb=False, logger=None):
    """List the dumped files, uploaded sources, or hubdb content."""
    logger = logger or get_logger(__name__)
    if dump is False and upload is False and hubdb is False:
        # if all set to False, we list both dump and upload as the default
        dump = upload = True

    _plugin = load_plugin(plugin_name, dumper=True, uploader=False, logger=logger)
    if dump:
        data_folder = _plugin.dumper.current_data_folder
        if not data_folder:
            # data_folder should be saved in hubdb already, if dump has been done successfully first
            logger.error('Data folder is not available. Please run "dump" first.')
            # Typically we should not need to use new_data_folder as the data_folder,
            # but we keep the code commented out below for future reference
            # utils.run_sync_or_async_job(dumper.create_todump_list, force=True)
            # data_folder = dumper.new_data_folder
        else:
            show_dumped_files(data_folder, _plugin.plugin_name)
    if upload:
        show_uploaded_sources(pathlib.Path(_plugin.data_plugin_dir), _plugin.plugin_name)
    if hubdb:
        show_hubdb_content()


########################
# for serve command    #
########################


def serve(host, port, plugin_name, table_space):
    """Serve a simple API server to query the data plugin source."""
    from .web_app import main

    src_db = get_src_db()
    rprint(f"[green]Serving data plugin source: {plugin_name}[/green]")
    asyncio.run(main(host=host, port=port, db=src_db, table_space=table_space))


def do_serve(plugin_name=None, host="localhost", port=9999, logger=None):
    logger = logger or get_logger(__name__)
    _plugin = load_plugin(plugin_name, dumper=False, uploader=True, logger=logger)
    uploader_classes = _plugin.uploader_classes
    table_space = [item.name for item in uploader_classes]
    serve(host=host, port=port, plugin_name=_plugin.plugin_name, table_space=table_space)


########################
# for clean command    #
########################


def remove_files_in_folder(folder_path):
    """Remove all files in a folder."""
    for filename in os.listdir(folder_path):
        file_path = os.path.join(folder_path, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            rprint("[red]Failed to delete %s. Reason: %s [/red]" % (file_path, e))
    shutil.rmtree(folder_path)


def do_clean_dumped_files(data_folder, plugin_name):
    """Remove all dumped files by a data plugin in the data folder."""
    if not os.path.isdir(data_folder):
        rprint(f"[red]Data folder {data_folder} not found! Nothing has been dumped yet[/red]")
        return
    if not os.listdir(data_folder):
        rprint("[red]Empty folder![/red]")
    else:
        rprint(f"[green]There are all files dumped by [bold]{plugin_name}[/bold]:[/green]")
        print("\n".join(os.listdir(data_folder)))
        delete = typer.confirm("Do you want to delete them?")
        if not delete:
            raise typer.Abort()
        remove_files_in_folder(data_folder)
        rprint("[green]Deleted![/green]")


def do_clean_uploaded_sources(working_dir, plugin_name):
    """Remove all uploaded sources by a data plugin in the working directory."""
    uploaders = get_uploaders(working_dir)
    src_db = get_src_db()
    uploaded_sources = []
    for item in src_db.collection_names():
        if item in uploaders:
            uploaded_sources.append(item)
        for uploader_name in uploaders:
            if item.startswith(f"{uploader_name}_archive_") or item.startswith(f"{uploader_name}_temp_"):
                uploaded_sources.append(item)
    if not uploaded_sources:
        rprint("[red]No source has been uploaded yet! [/red]")
    else:
        rprint(f"[green]There are all sources uploaded by [bold]{plugin_name}[/bold]:[/green]")
        print("\n".join(uploaded_sources))
        delete = typer.confirm("Do you want to drop them?")
        if not delete:
            raise typer.Abort()
        for source in uploaded_sources:
            src_db[source].drop()
        rprint("[green]All collections are dropped![/green]")


def do_clean(plugin_name=None, dump=False, upload=False, clean_all=False, logger=None):
    """Clean the dumped files, uploaded sources, or both."""
    logger = logger or get_logger(__name__)
    if clean_all:
        dump = upload = True
    if dump is False and upload is False:
        logger.error("Please provide at least one of following option: --dump, --upload, --all")
        raise typer.Exit(1)

    _plugin = load_plugin(plugin_name, dumper=True, uploader=False, logger=logger)

    if dump:
        data_folder = _plugin.dumper.current_data_folder
        if not data_folder:
            # data_folder should be saved in hubdb already, if dump has been done successfully first
            logger.error('Data folder is not available. Please run "dump" first.')
        do_clean_dumped_files(data_folder, _plugin.plugin_name)
    if upload:
        do_clean_uploaded_sources(_plugin.data_plugin_dir, _plugin.plugin_name)



"""
Request Argument Standardization
"""

import logging
import re
from collections import UserDict, abc, defaultdict
from datetime import datetime as dt
from pprint import pformat
from types import MappingProxyType

import jmespath
import orjson

try:
    from re import Pattern  # py>=3.7
except ImportError:
    from typing import Pattern

try:
    from functools import singledispatchmethod  # py>=3.7
except ImportError:
    from singledispatchmethod import singledispatchmethod

from biothings.utils.common import dotdict, split_ids


class OptionError(ValueError):
    def __init__(self, reason=None, **kwargs):
        super().__init__()
        self.info = {"reason": reason}
        self.info.update(kwargs)

    def simplify(self):
        self.info = {k: v for k, v in self.info.items() if v}

    def __str__(self):
        return f"OptionError({pformat(self.info)})"


class Converter:
    """
    A generic HTTP request argument processing unit.
    Only perform one level of validation at this moment.
    The strict switch controls the type conversion rules.
    """

    def __init__(self, **kwargs):
        self.keyword = kwargs.get("keyword")
        self.type_ = kwargs.get("type", str)
        self.strict = kwargs.get("strict", True)
        translations = kwargs.get("translations", ())

        # Opinion on what value self.type_ can take
        # ---------------------------------------------
        # Although it may sound attractive to use non-native
        # "type" here to simplify certain object constuction
        # through the converstion process, and indeed it is
        # supported through self.to_type, this could lead to
        # difficulties in OptionSet serialization and other
        # problems in integration with this module.

        self.translations = []

        if isinstance(translations, dict):
            translations = translations.items()

        for pattern, repl in translations:
            if isinstance(pattern, Pattern):
                self.translations.append((pattern, repl))
            elif isinstance(pattern, tuple):
                self.translations.append((re.compile(*pattern), repl))
            elif isinstance(pattern, str):
                self.translations.append((re.compile(pattern), repl))
            else:  # https://docs.python.org/3/library/re.html#re.compile
                raise TypeError("Invalid Regex Pattern.")

    def __call__(self, value, to_type):
        return self.convert_to(value, to_type)

    @classmethod
    def subclasses(cls, kwargs):
        for kls in cls.__subclasses__():
            name = kls.__name__[: -len("ArgCvter")]
            yield name.lower(), kls(**kwargs)

    def convert(self, value):
        return self.convert_to(value, self.type_)

    def convert_to(self, value, to_type):
        # default implementation
        # only works for strings
        assert isinstance(value, str)

        if to_type is None:
            to_type = self.type_

        if to_type is str:
            return value  # pass through

        if to_type is bool:
            return self.str_to_bool(value)

        if to_type is int:
            return self.str_to_int(value)

        if to_type in (list, tuple, set):
            lst = self.str_to_list(value)
            return self.to_type(lst, to_type)

        return self.to_type(value, to_type)

    def translate(self, value):
        if isinstance(value, (tuple, list, set)):
            return (type(value))(self.translate(item) for item in value)

        if not isinstance(value, str):
            return value  # can only perform translations to strings

        # https://docs.python.org/3/library/re.html#re.sub
        for pattern, repl in self.translations:
            value = re.sub(pattern, repl, value)

        if self.keyword == "jmespath" and value:
            # processing jmespath parameter to be a tuple of (parent_path, target_field, jmes_query)
            try:
                target_field_path, jmes_query = value.split("|", maxsplit=1)
                jmes_query = jmespath.compile(jmes_query)
            except ValueError as err:  # JMES exeptions are subclasses of ValueError
                raise OptionError(keyword=self.keyword, reason="Invalid value for jmespath parameter", details=str(err))
            # now split target_field_path into parent_path and target_field
            target_field_path = target_field_path or "."  # set to root field if not provided
            try:
                parent_path, target_field = target_field_path.rsplit(".", maxsplit=1)
            except ValueError:
                # if no . in the path, it means the target field is the root field
                parent_path, target_field = "", target_field_path
            value = parent_path, target_field, jmes_query

        return value

    @staticmethod
    def str_to_bool(val):
        """Interpret string representation of bool values."""
        assert isinstance(val, str)
        try:  # if it is a number
            return float(val) > 0
        except ValueError:
            pass  # process as keywords
        return val.lower() in ("1", "true", "yes", "y", "t")

    # Opinion on str -> list
    #
    # It appears to have become more problematic recently, as the variety
    # of data increase in biothings applications, causing the identifiers
    # and field values hard to escape properly when used in queries.
    #
    # Consider implementing a very safe splitting algorithm that only works
    # for basic cases like a,b,c and use other methods like JSON input to
    # pass in complex queries.
    #
    # For example, for field reagent.GNF_mm+hs-MGC in Mygene:
    # the current algorithm will split it into ['reagent.GNF_mm', 'hs-MGC'

    @staticmethod
    def str_to_list(val):
        """Cast Biothings-style str to list."""
        try:  # core splitting algorithm
            lst = split_ids(str(val))
        except ValueError as err:
            raise OptionError(str(err))
        return lst

    def str_to_int(self, val):
        """Convert a numerical string to an integer."""
        assert isinstance(val, str)
        if not self.strict:
            val = self.to_type(val, float)
        return self.to_type(val, int)

    @staticmethod
    def to_type(val, type_):
        """
        Native type casting in Python.
        Fallback approach for type casting.
        """
        try:
            result = (type_)(val)
        except (ValueError, TypeError):
            raise OptionError(f"Expect type {type_.__name__}.")
        return result


class PathArgCvter(Converter):
    """
    Dedicated argument converter for path arguments.
    Correspond to arguments received in tornado for
        RequestHandler.path_args
        RequestHandler.path_kwargs
    See https://www.tornadoweb.org/en/stable/web.html
    """


class QueryArgCvter(Converter):
    """
    Dedicated argument converter for url query arguments.
    Correspond to arguments received in tornado from
        RequestHandler.get_query_argument
    See https://www.tornadoweb.org/en/stable/web.html
    """

    @classmethod
    def str_to_bool(cls, val):
        """Biothings-style str to bool interpretation"""
        # empty string indicates the presence of its key
        # we consider this to be a positive boolean value
        # this is especially useful in url when the user
        # may specify endpoint?op1&op2&op3 in which case
        # it makes sense to consider their keys true.
        return super().str_to_bool(val) or val.lower() == ""


class FormArgCvter(Converter):
    """
    Dedicated argument converter for HTTP body arguments.
    Additionally support JSON seriealization format as values.
    Correspond to arguments received in tornado from
        RequestHandler.get_body_argument
    See https://www.tornadoweb.org/en/stable/web.html
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # If jsoninput evaluates to true,
        # try to load the str as a json dump
        jsoninput = kwargs.get("jsoninput", False)
        if isinstance(jsoninput, bool):
            self.jsoninput = jsoninput
        else:  # it itself can be in argument format
            self.jsoninput = self.convert_to(jsoninput, bool)

    # Opinion on JsonInput
    #
    # The aforementioned feature was partly for convenience and partly
    # for backward compatibility with the original published design.
    #
    # It might be more beneficial to use standard data serialization
    # format like JSON to indicate data type unequivocally.

    def convert_to(self, value, to_type):
        if self.jsoninput:
            try:  # attempt to load as json first
                _value = orjson.loads(value)
            except orjson.JSONDecodeError as exc:
                logging.debug(repr(exc))
            else:  # no more conversions
                if isinstance(_value, to_type):
                    return _value
        if isinstance(value, to_type):
            return value
        if isinstance(value, str):
            return super().convert_to(value, to_type)
        return self.to_type(value, to_type)


class JsonArgCvter(Converter):
    """
    Dedicated argument converter for JSON HTTP bodys.
    Here it is used for dict JSON objects, with their
    first level keys considered as parameters and
    their values considered as arguments to process.

    May correspond to this tornado implementation:
    https://www.tornadoweb.org/en/stable/web.html#input
    """

    def convert_to(self, value, to_type):
        if isinstance(value, to_type):
            return value  # type matches

        if self.strict:
            # since JSON support value types
            # strict mode enforces it and essentially
            # makes this step a validation step.
            raise OptionError(f"Expect type {to_type.__name__}.")

        # mostly for biothings client 0.2.3 compatibility
        if isinstance(value, str) and to_type is list:
            return self.str_to_list(value)

        return self.to_type(value, to_type)

    def to_type(self, val, type_):
        if issubclass(type_, (list, tuple, set)) and not self.strict:
            val = (val,)  # "abc" -> ["abc"] instead of ["a", "b", "c"]

        return super().to_type(val, type_)


class ReqResult(dotdict):
    # besides multi-level attribute(dot)
    # read and write access, this class
    # also returns None value for missing
    # keys instead of raising an error

    def __str__(self):
        return f"ReqResult({pformat(self)})"


class ReqArgs:
    class Path:
        def __init__(self, args=None, kwargs=None):
            if args is None:
                args = ()
            if kwargs is None:
                kwargs = {}
            assert isinstance(args, (tuple, list))
            assert isinstance(kwargs, dict)
            self.args = args
            self.kwargs = kwargs

        def __getitem__(self, key):
            try:
                if isinstance(key, int):
                    return self.args[key]
                if isinstance(key, str):
                    return self.kwargs[key]
            except (KeyError, IndexError):
                return None

        def __bool__(self):
            return bool(self.args) or bool(self.kwargs)

        def __str__(self):
            lines = []
            for src in ("args", "kwargs"):
                if getattr(self, src):
                    lines.append("{}={}".format(src, str(getattr(self, src))))
            return "Path(" + ", ".join(lines) + (")")

    def __init__(self, path=None, query=None, form=None, json_=None):
        assert isinstance(query, (dict, type(None)))
        assert isinstance(form, (dict, type(None)))

        if not isinstance(path, (self.Path, type(None))):
            path = self.Path(*path)

        self.path = path  # positional and named capture group in a routing pattern
        self.query = query  # key value pairs after a question mark at the end of an url
        self.form = form  # type multipart/form-data and application/x-www-form-urlencoded
        self.json = json_ if isinstance(json_, dict) else {}  # type application/json

    def lookup(self, locator, order=None, src=False):
        if isinstance(locator, str):
            locator = Locator(dict(keyword=locator))
        elif isinstance(locator, dict):
            locator = Locator(locator)
        elif not isinstance(locator, Locator):
            raise TypeError("Unknown Locator.")

        if order is None:
            order = ("path", "query", "form", "json")
        elif isinstance(order, str):
            order = (order,)
        elif not isinstance(order, abc.Iterable):
            raise TypeError("Unknown Order.")

        for loc in order:
            try:
                args = getattr(self, loc)
                val = locator.lookin(args)
            except AttributeError:
                _ = "No such location: %s."
                logging.warning(_, loc)
            else:
                if val is not None:
                    return (val, loc) if src else val

        return (None, None) if src else None

    def __str__(self):
        lines = []
        for src in ("path", "query", "form", "json"):
            if getattr(self, src):
                lines.append("{}={}".format(src, str(getattr(self, src))))
        return "ReqArgs(" + ",\n".join(lines) + (")")


class Locator:
    """
    Describes the location of an argument in ReqArgs.
    {
        "keyword": <str>,
        "path": <int or str>,
        "alias": <str or [<str>, ...]>
    }
    """

    def __init__(self, defdict):
        self.keyword = defdict.get("keyword")
        self.path = defdict.get("path")
        aliases = defdict.get("alias", [])

        assert isinstance(self.path, (str, int, type(None)))
        assert isinstance(self.keyword, (str, type(None)))

        if isinstance(aliases, (list, tuple)):
            self.aliases = aliases
        elif isinstance(aliases, str):
            self.aliases = [aliases]
        else:  # validation failed
            raise ValueError("Unknown Alias.")

    @singledispatchmethod
    def lookin(self, location):
        """
        Find an argument in the specified location.
        Use directions indicated in this locator.
        """

    @lookin.register(ReqArgs.Path)
    def _(self, path):
        if self.path is not None:
            return path[self.path]
        return None

    @lookin.register(dict)  # all others
    def _(self, dic):
        if self.keyword in dic:
            return dic[self.keyword]
        for alias in self.aliases:
            if alias in dic:
                return dic[alias]
        return None


class Existentialist:
    """
    Describes the requirement of
    the existance of an argument.
    {
        "default": <object>,
        "required": <bool>,
    }
    """

    def __init__(self, defdict):
        self._defdict = MappingProxyType(defdict)
        self.keyword = defdict.get("keyword")
        self.required = bool(defdict.get("required"))
        self.default = defdict.get("default")

        if self.default and self.required:
            logging.warning(
                (
                    "A default value is set for parameter '%s' "
                    "while 'required' is set to True, making it"
                    "ineffective."
                ),
                self.keyword,
            )

    def inquire(self, obj):
        if obj is None:
            if self.required:
                raise OptionError(
                    missing=self.keyword,
                    keyword=None,  # empty this field
                    alias=self._defdict.get("alias"),
                )

            obj = self.default

        return obj


class Validator:
    """
    Describes the requirement of
    the existance of an argument.
    {
        "enum": <container>,
        "max": <int>,
        "min": <int>,
        "date_format": <str>,
    }
    """

    def __init__(self, defdict):
        self._defdict = MappingProxyType(defdict)
        self.keyword = defdict.get("keyword")
        self.strict = defdict.get("strict", True)
        self.enum = defdict.get("enum", ())
        self.max = defdict.get("max")
        self.min = defdict.get("min")
        self.date_format = defdict.get("date_format")

        assert isinstance(self.enum, abc.Container)
        assert isinstance(self.max, (int, type(None)))
        assert isinstance(self.min, (int, type(None)))
        assert isinstance(self.date_format, (str, type(None)))

    def validate(self, obj):
        if self.enum and not self._in_enum(obj):
            raise OptionError(
                keyword=self.keyword,
                allowed=self.enum,
                alias=self._defdict.get("alias"),
            )

        if self.max:
            if isinstance(obj, (list, tuple, set)):
                self._check_list_max(obj)
            elif isinstance(obj, (int, float, complex)):
                self._check_num_max(obj)

        if self.min:
            if isinstance(obj, (list, tuple, set)):
                self._check_list_min(obj)
            elif isinstance(obj, (int, float, complex)):
                self._check_num_min(obj)

        if self.date_format and isinstance(obj, str):
            self._check_date_format(obj)
        return obj

    def _in_enum(self, value):
        if isinstance(value, (list, tuple, set)):
            for val in value:
                if not self._in_enum(val):
                    return False
            return True

        return value in self.enum

    def _check_list_max(self, container):
        if len(container) > self.max:
            raise OptionError(keyword=self.keyword, max=self.max, size=len(container))

    def _check_num_max(self, num):
        if isinstance(num, bool):
            return

        if num > self.max:
            raise OptionError(keyword=self.keyword, max=self.max, num=num)

    def _check_list_min(self, container):
        if len(container) > self.min:
            raise OptionError(keyword=self.keyword, min=self.min, size=len(container))

    def _check_num_min(self, num):
        if isinstance(num, bool):
            return

        if num < self.min:
            raise OptionError(keyword=self.keyword, min=self.min, num=num)

    def _check_date_format(self, value):
        try:
            dt.strptime(value, self.date_format)
        except Exception:
            raise OptionError(keyword=self.keyword, date_format=self.date_format, num=value)


# For Future Work
# ------------------

# Consider supporting the OpenAPI-compatible JSON Schema
# definitions as a subset of keys to describe the Option.
# By doing this, it is both easier to validate the input
# and to generate an OpenAPI Specification. Validation
# can be performed by existing packages like "jsonschema",
# and components of the OpenAPI Specifications can be
# generated by directly taking the relevant keys from the
# Option definition. However, keep in mind that validation
# and converstion will still not be straightforward when
# dealing with query parameters and form-encoded requests.

# To illustrate the benefit of using a subset of JSON Schema
# as our option validation language, take the example of
# validating complex objects. Currently, nested objects or
# containers are only validated at the top level, with the
# help of JSON schema, we could easily find the syntax needed
# to define the object structure and perform validation.
# Comparing to designing the schema language ourselves,
# and writing validation code for that purpose, using an
# existing framework should save time and improve accuracy,
# not to mention the additional benefit of spec generation.

# For more about OpenAPI-compatible JSON Schema:
# https://swagger.io/specification/#schema-object


class Option(UserDict):
    """
    A parameter for end applications to consume.
    Find the value of it in the desired *location*.

    For example:
    {
        "keyword": "q",
        "location": ("query", "form", "json"),
        "default": "__all__",
        "type": "str"
    }
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._locater = Locator(self)
        self._exists = Existentialist(self)
        self._typify = dict(Converter.subclasses(self))
        self._validator = Validator(self)

        # default argument parsing location order
        self.order = self.get("location", None)
        if self.order == "body":  # shortcut
            self.order = ("form", "json")

    def parse(self, reqargs):
        if not isinstance(reqargs, ReqArgs):
            reqargs = ReqArgs(*reqargs)

        # ------- debug here -------
        # if self.get("keyword") == "q":
        #     print()  # breakpoint
        # ---------------------------

        # find the user input
        val, loc = reqargs.lookup(locator=self._locater, order=self.order, src=True)

        if val is None:
            val = self._exists.inquire(val)
        else:  # type conversion and transform
            val = self._typify[loc].convert(val)
            val = self._typify[loc].translate(val)

        # additional conditions
        val = self._validator.validate(val)

        return val


class OptionSet(UserDict):
    """
    A collection of options that a specific endpoint consumes.
    Divided into *groups* and by the *request methods*.

    For example:
    {
        "*":{"raw":{...},"size":{...},"dotfield":{...}},
        "GET":{"q":{...},"from":{...},"sort":{...}},
        "POST":{"q":{...},"scopes":{...}}
    }
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.groups = set()  # explicit result access groups
        self.setup()  # populate self.optset variable

    def __missing__(self, key):
        self.data[key] = {}
        return self[key]

    def setup(self):
        """
        Apply the wildcard method configurations dict.
        Must call this method after changes to this object.
        """
        # store Option objects used for parse method
        self.optset = defaultdict(dict)

        for method, options in self.data.items():
            for keyword, defdict in options.items():
                option = dict(defdict)
                option["keyword"] = keyword
                self.optset[method][keyword] = Option(option)
                if "group" in option:
                    if isinstance(option["group"], (list, tuple, set)):
                        self.groups.update(option["group"])
                    elif isinstance(option["group"], str):
                        self.groups.add(option["group"])

        wildcards = self.optset.get("*", {})
        for _method, options in self.optset.items():
            for keyword, option in wildcards.items():
                options.setdefault(keyword, option)

    def parse(self, method, reqargs):
        """
        Parse a HTTP request, represented by its method and args,
        with this OptionSet and return an attribute dictionary.
        """

        options = self.optset.get(method, self.optset["*"])
        result = defaultdict(dict)  # to accomodate groups

        for keyword, option in options.items():
            try:
                val = option.parse(reqargs)
            except OptionError as err:
                err.info.setdefault("keyword", keyword)
                err.info["alias"] = option.get("alias")
                err.simplify()  # remove empty fields
                raise err  # with helpful info

            if val is not None:
                if "group" in option:
                    group = option["group"]
                    if isinstance(group, str):
                        result[group][keyword] = val
                    else:  # assume iterable
                        for _group in group:
                            result[_group][keyword] = val
                else:  # top level keywords
                    result[keyword] = val
            elif "default" in option:  # explicit None
                result[keyword] = None

        # make sure all named groups exist
        for group in self.groups:
            if group not in result:
                result[group] = {}

        return ReqResult(result)


class OptionsManager(UserDict):
    """
    A collection of OptionSet(s) that makes up an application.
    Provide an interface to setup and serialize.

    Example:
    {
        "annotation": {"*": {...}, "GET": {...}, "POST": {... }},
        "query": {"*": {...},  "GET": {...}, "POST": {... }},
        "metadata": {"GET": {...}, "POST": {... }}
    }
    """

    def add(self, name, optionset, groups=()):
        if not name:
            logging.warning("Ignore unnamed optionset:\n%s", optionset)
        if name not in self.data:
            self.data[name] = OptionSet(optionset)
            self.data[name].groups.update(groups)
        else:  # update existing optionset
            for method, options in optionset.items():
                # merge second level objects
                self.data[name][method].update(options)
                self.data[name].groups.update(groups)
                self.data[name].setup()  # required

    def log(self):
        # serializable API-ready format
        return self._serialize(self)

    def _serialize(self, obj):
        if isinstance(obj, abc.Mapping):
            _obj = {}
            items = list(obj.items())
            for key, val in sorted(items):
                _obj[key] = self._serialize(val)
            return _obj
        if isinstance(obj, (list, tuple)):
            return [self._serialize(item) for item in obj]
        if isinstance(obj, (str, int)):
            return obj
        if hasattr(obj, "__name__"):
            return obj.__name__
        return str(obj)  # best effort
